import os
import sys
import wiz
import re
import traceback
import time
import taurus.qt.qtgui.resource
import pool_editor_UI
import pool_editor_UI
import ms_editor_UI
from PyQt4 import QtGui, QtCore, Qt
from taurus.core.util.enumeration import Enumeration
from sardana.taurus.core.tango.sardana import SardanaManager

"""This wizard is designed for creating new Sardana.

The wizard consist:

- TangoPage
- SardanaPage
- PoolPage
- MSPage
- CommitPage
- OutroPage

"""


class NewSardanaIntroPage(wiz.SardanaIntroBasePage):
    """
    Introduction page
    """

    def __init__(self, parent=None):
        wiz.SardanaIntroBasePage.__init__(self, parent)
        self.sardana_name = None

    def getIntroText(self):
        return """This wizard will guide you through the creation of the new instance of Sardana"""

    def setNextPageId(self, id):
        self._nextPageId = id


class SelectTangoHostPage(wiz.SardanaBasePage):
    """
    Page for selecting tango host
    """

    def __init__(self, parent=None):
        wiz.SardanaBasePage.__init__(self, parent)

        self.setSubTitle('Please select the Tango Host')
        self._valid = False

        self._regExp0 = QtCore.QRegExp("^[0-9a-zA-Z]{,50}")
        self._regValid0 = QtGui.QRegExpValidator(self._regExp0, self)

        self._regExp1 = QtCore.QRegExp("^[0-9]{,50}")
        self._regValid1 = QtGui.QRegExpValidator(self._regExp1, self)

        self._panel = self.getPanelWidget()

        layout = QtGui.QFormLayout()
        self.hostEdit = QtGui.QLineEdit()
        self.hostEdit.setValidator(self._regValid0)

        self.portEdit = QtGui.QLineEdit()
        self.portEdit.setValidator(self._regValid1)
        self.checkButton = QtGui.QPushButton("Check")

        layout.addRow("&Host", self.hostEdit)
        layout.addRow("&Port", self.portEdit)
        layout.addRow("", self.checkButton)
        self.checkButton.hide()
        host, port = self.get_default_tango_host()

        self.hostEdit.setText(host)
        self.portEdit.setText(port)

        self._panel.setLayout(layout)

        self['db'] = self._getDatabase
        self['port'] = self._getPort
        self['host'] = self._getHost

        self.connect(self.hostEdit, QtCore.SIGNAL(
            'textEdited(const QString &)'), self._letterChanged)
        self.connect(self.portEdit, QtCore.SIGNAL(
            'textEdited(const QString &)'), self._letterChanged)
        self.connect(self.checkButton, QtCore.SIGNAL(
            'clicked()'), self.checkData)

        self.checkData()

    def get_default_tango_host(self):
        tg = os.environ.get('TANGO_HOST', '')
        if not tg or tg.count(':') != 1:
            return '', ''
        return tg.split(':')

    def _letterChanged(self):
        self._valid = False
        self.setStatus('please click the ''Check'' button')
        self.emit(QtCore.SIGNAL('completeChanged()'))
        self.checkButton.show()

    def checkData(self):
        try:
            db = self.db()
            if not db is None:
                self.setStatus('please click the ''Next'' button')
                self._valid = True
                self.emit(QtCore.SIGNAL('completeChanged()'))
        except Exception, e:
            self.setStatus('Invalid database')
            self._valid = False
        self.checkButton.hide()

    def _getPort(self):
        return str(self.portEdit.text()).lower()

    def _getHost(self):
        return str(self.hostEdit.text()).lower()

    def _getDatabase(self):
        host = str(self.hostEdit.text()).lower()
        port = str(self.portEdit.text())

        return taurus.Database("%s:%s" % (host, port))

    def host(self):
        return self.wizard()['host']

    def port(self):
        return self.wizard()['port']

    def db(self):
        return self.wizard()['db']

    def isComplete(self):
        return self._valid

    def setNextPageId(self, id):
        self._nextPageId = id

    def nextId(self):
        return self._nextPageId

    def initializePage(self):
        wiz.SardanaBasePage.initializePage(self)
        self.checkData()


class AddSardanaBasePage(wiz.SardanaBasePage):
    """
    Page for setting Sardana name
    """

    def __init__(self, parent=None):
        wiz.SardanaBasePage.__init__(self, parent)
        self.setSubTitle('Please type the name for the new Sardana instance')
        self._valid = False

        panel = self.getPanelWidget()

        layout = QtGui.QFormLayout()

        self.nameEdit = QtGui.QLineEdit()
        self.nameEdit.setText("MySardana")

        self._regExp = QtCore.QRegExp("^[0-9a-zA-Z]{,50}")
        self._regValid = QtGui.QRegExpValidator(self._regExp, self)

        self.nameEdit.setValidator(self._regValid)

        self.checkButton = QtGui.QPushButton("Check")
        self.checkButton.hide()

        layout.addRow("&Name", self.nameEdit)
        layout.addRow("", self.checkButton)

        panel.setLayout(layout)

        self.connect(self.nameEdit, QtCore.SIGNAL(
            'textEdited(const QString &)'), self._letterChanged)
        self.connect(self.checkButton, QtCore.SIGNAL(
            'clicked()'), self.checkData)

    def _getSardana(self):
        sardana = str(self.nameEdit.text())
        return sardana

    def _letterChanged(self):
        self._valid = False
        self.emit(QtCore.SIGNAL('completeChanged()'))
        self.setStatus('please click the ''Check'' button')
        self.checkButton.show()

    def sardana(self):
        return self.wizard()['sardana']

    def db(self):
        return self.wizard()['db']

    def checkData(self):
        if self.nameEdit.text() == "":
            self.setStatus('Please type the name')
            self._valid = False
        else:
            if self.nameEdit.text() not in self._getSardanas():
                self.setStatus('please click the ''Next'' button')
                self._valid = True
                self.emit(QtCore.SIGNAL('completeChanged()'))
            else:
                self.setStatus('The name already exists in the database')
                self._valid = False
        self.checkButton.hide()

    def isComplete(self):
        return self._valid

    def _getSardanas(self):
        db = self.db()
        services = db.get_service_list('Sardana.*')
        service_instances = []
        for service in services:
            service_instances.append(service.split('/', 1)[1])
        return service_instances

    def initializePage(self):
        wiz.SardanaBasePage.initializePage(self)
        self.wizard().__setitem__("sardana", self._getSardana)
        self.checkData()

    def setNextPageId(self, id):
        self._nextPageId = id

    def nextId(self):
        return self._nextPageId


class Item(object):
    """
    SuperClass of Pool and Macroservers items
    """

    def __init__(self, instanceName="", hostName="", level=None):
        self._instanceName = str(instanceName)
        self._hostName = str(hostName)
        self._level = level

    def setInstanceName(self, name):
        self._instanceName = str(name)

    def getInstanceName(self):
        return self._instanceName

    def setHostName(self, name):
        self._hostName = str(name)

    def getHostName(self):
        return self._hostName

    def setLevel(self, level):
        self._level = level

    def getLevel(self):
        return self._level


class SimpleEditorBasePage(wiz.SardanaBasePage):
    """
    Superclass of Pool and Macroserver editor
    """

    def __init__(self, parent=None):
        wiz.SardanaBasePage.__init__(self, parent)
        self.setSubTitle('SubTitle')
        panel = self.getPanelWidget()
        self._valid = True
        self.horizontalLayout = QtGui.QHBoxLayout(panel)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.tableWidget = QtGui.QTableWidget(self)
        self.tableWidget.setObjectName("tableWidget")
        self.tableWidget.setColumnCount(3)
        self.tableWidget.setRowCount(0)
        self.tableWidget.setSelectionMode(
            QtGui.QAbstractItemView.SingleSelection)
        self.tableWidget.setSelectionBehavior(
            QtGui.QAbstractItemView.SelectRows)
        self.horizontalLayout.addWidget(self.tableWidget)
        self._setTableHeaders()
        self.tableWidget.horizontalHeader().resizeSection(0, 125)
        self.tableWidget.horizontalHeader().resizeSection(1, 120)
        self.tableWidget.horizontalHeader().resizeSection(2, 55)
        self.tableWidget.verticalHeader().setVisible(False)

        self.verticalLayout = QtGui.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")

        self.addButton = QtGui.QPushButton(self)
        self.addButton.setText(QtGui.QApplication.translate(
            "Form", "Add", None, QtGui.QApplication.UnicodeUTF8))
        self.addButton.setObjectName("addButton")
        self.addButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-add"))
        self.verticalLayout.addWidget(self.addButton)

        self.editButton = QtGui.QPushButton(self)
        self.editButton.setText(QtGui.QApplication.translate(
            "Form", "Edit", None, QtGui.QApplication.UnicodeUTF8))
        self.editButton.setObjectName("editButton")
        self.editButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("preferences-system"))
        self.verticalLayout.addWidget(self.editButton)

        self.removeButton = QtGui.QPushButton(self)
        self.removeButton.setText(QtGui.QApplication.translate(
            "Form", "Remove", None, QtGui.QApplication.UnicodeUTF8))
        self.removeButton.setObjectName("removeButton")
        self.removeButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-remove"))
        self.verticalLayout.addWidget(self.removeButton)

        spacerItem = QtGui.QSpacerItem(
            20, 40, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.horizontalLayout.addLayout(self.verticalLayout)
        #spacerItem1 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        # self.horizontalLayout.addItem(spacerItem1)
        self.listOfItems = []
        self._comboList = []
        self._spinList = []

    def _setTableHeaders(self):
        self.tableWidget.setColumnCount(3)
        self.tableWidget.setHorizontalHeaderItem(
            0, QtGui.QTableWidgetItem("name"))
        self.tableWidget.setHorizontalHeaderItem(
            1, QtGui.QTableWidgetItem("host"))
        self.tableWidget.setHorizontalHeaderItem(
            2, QtGui.QTableWidgetItem("level"))

    def openEditor(self, item=None):
        pass

    def checkName(self, name, id=None):
        i = 0
        for item in self.listOfItems:
            if i != id:
                if name == item.getInstanceName():
                    return False
            i += 1
        return True

    def addItem(self, item):
        if item.getLevel() is None:
            item.setLevel("1")
        item.setHostName(self._getHostList()[0])
        self.listOfItems.append(item)
        self._fillList()
        self.checkData()
        return True

    def editItem(self, item, id):
        if item.getLevel() is None:
            item.setLevel("1")
        self.listOfItems[id] = item
        self._fillList()

    def removeItem(self, id):
        self.listOfItems.remove(self.listOfItems[id])
        self._fillList()
        self.checkData()
        return True

    def hostChanged(self, hostId=None):
        name = str(self.sender().objectName())
        id = int(name[name.index('_') + 1:])
        if self.sender().currentText() == "None":
            self.listOfItems[id].setHostName("")
        else:
            self.listOfItems[id].setHostName(self.sender().currentText())
        self._fillList()

    def levelChanged(self, id=None):
        name = str(self.sender().objectName())
        id = int(name[name.index('_') + 1:])
        self.listOfItems[id].setLevel(str(self.sender().value()))

    def checkData(self):
        self._valid = True

    def _getItemList(self):
        return self.listOfItems[:]

    def isComplete(self):
        return self._valid

    def _fillList(self):
        self._comboList = []
        self._spinList = []
        self.tableWidget.clear()
        self._setTableHeaders()

        self.tableWidget.setRowCount(len(self.listOfItems))
        self.tableWidget.clearSelection()
        # self.tableWidget.setItemSelected(None)

        # self.tableWidget.selectRow(-1)

        self.editButton.setObjectName("editButton")
        i = 0
        for item in self.listOfItems:
            self.tableWidget.setCellWidget(i, 0, QtGui.QLabel(item.text()))
            self._comboList.append(QtGui.QComboBox())
            self._comboList[i].setObjectName("hostComboBox_%d" % i)
            if item.getHostName() == "":
                self._comboList[i].addItem("None")
            else:
                self._comboList[i].addItem(item.getHostName())
            self._comboList[i].addItems(self._hostList)
            self._comboList[i].addItem("None")
            self._comboList[i].insertSeparator(len(self._hostList) + 1)
            QtCore.QObject.connect(self._comboList[i], QtCore.SIGNAL(
                "currentIndexChanged(int)"), self.hostChanged)
            self.tableWidget.setCellWidget(i, 1, self._comboList[i])
            self._spinList.append(QtGui.QSpinBox())
            self._spinList[i].setRange(*SardanaManager().get_level_range())
            self._spinList[i].setObjectName("levelSpinBox_%d" % i)
            QtCore.QObject.connect(self._spinList[i], QtCore.SIGNAL(
                "valueChanged(int)"), self.levelChanged)
            self.tableWidget.setCellWidget(i, 2, self._spinList[i])

            if (item.getHostName() == "") or (item.getHostName() is None):
                self._spinList[i].setDisabled(True)
                self._spinList[i].setValue(0)
            else:
                self._spinList[i].setValue(int(item.getLevel()))
            i += 1

    def initializePage(self):
        wiz.SardanaBasePage.initializePage(self)
        self._hostList = self._getHostList()
        self._fillList()
        self.checkData()

    def setNextPageId(self, id):
        self._nextPageId = id

    def db(self):
        return self.wizard()['db']

    def _getHostList(self):
        return SardanaManager().get_hosts()

    def getPoolServerList(self):
        return [s[s.index('/') + 1:] for s in self.db().get_server_list() if s.startswith('Pool/')]


class Pool(Item):
    """
    Represents a Pool
    """

    def __init__(self, instanceName="", hostName="", level=None, poolDeviceName="", alias="", poolVersion="", poolPath=[]):

        Item.__init__(self, instanceName, hostName, level)
        self._poolDeviceName = str(poolDeviceName)
        self._alias = str(alias)
        self._poolVersion = str(poolVersion)
        self._poolPath = map(str, poolPath)

    def setPoolDeviceName(self, name):
        self._poolDeviceName = str(name)

    def getPoolDeviceName(self):
        return str(self._poolDeviceName)

    def setAlias(self, alias):
        self._alias = str(alias)

    def getAlias(self):
        return self._alias

    def setPoolVersion(self, poolVersion):
        self._poolVersion = str(poolVersion)

    def getPoolVersion(self):
        return self._poolVersion

    def setPoolPath(self, path):
        self._poolPath = map(str, path)

    def getPoolPath(self):
        return self._poolPath

    def text(self):
        return str(self._instanceName)

    def copy(self):
        return Pool(self._instanceName, self._hostName, self._level, self._poolDeviceName, self._alias, self._poolVersion, self._poolPath[:])


class AddPoolBasePage(SimpleEditorBasePage):
    """
    Page for editing the list of Pools
    """

    def __init__(self, parent=None):
        self._editor = None
        SimpleEditorBasePage.__init__(self, parent)
        self._editor = PoolEditor(parent=self)
        self.selectedItem = None
        self.item_id = None
        self.setSubTitle(
            'You can use this manager if you would like to Add, Edit or Delete Pool entries in the database.')
        self.addButton.setText(QtGui.QApplication.translate(
            "Form", "Add Pool", None, QtGui.QApplication.UnicodeUTF8))
        self.editButton.setText(QtGui.QApplication.translate(
            "Form", "Edit Pool", None, QtGui.QApplication.UnicodeUTF8))
        self.removeButton.setText(QtGui.QApplication.translate(
            "Form", "Remove Pool", None, QtGui.QApplication.UnicodeUTF8))
        QtCore.QObject.connect(self.addButton, QtCore.SIGNAL(
            "clicked()"), self.openAddEditor)
        QtCore.QObject.connect(self.editButton, QtCore.SIGNAL(
            "clicked()"), self.openItemEditor)
        QtCore.QObject.connect(
            self.removeButton, QtCore.SIGNAL("clicked()"), self.delete)
        QtCore.QObject.connect(self.tableWidget, QtCore.SIGNAL(
            "cellDoubleClicked(int,int)"), self.openItemEditor)

    def delete(self):
        if len(self.tableWidget.selectedIndexes()) > 0:
            msgBox = QtGui.QMessageBox()
            msgBox.setText("The list of Pools has been modified.")
            self.item_id = self.tableWidget.selectedIndexes()[0].row()
            self.selectedItem = self.listOfItems[self.item_id]
            msgBox.setInformativeText(
                "Do you want to delete Pool?:\n" + self.selectedItem.text())
            msgBox.setStandardButtons(
                QtGui.QMessageBox().Ok | QtGui.QMessageBox().Cancel)
            msgBox.setDefaultButton(QtGui.QMessageBox().Cancel)
            msgBox.setIcon(QtGui.QMessageBox.Question)
            ret = msgBox.exec_()
            if ret == QtGui.QMessageBox().Ok:
                self.removeItem(self.item_id)
            if ret == QtGui.QMessageBox().Cancel:
                pass

    def openAddEditor(self):
        self.item_id = None
        self.selectedItem = None
        SimpleEditorBasePage.openEditor(self)
        self._editor.showEditor(item=None, item_id=None)

    def openItemEditor(self):
        if len(self.tableWidget.selectedIndexes()) > 0:
            self.item_id = self.tableWidget.selectedIndexes()[0].row()
            self.selectedItem = self.listOfItems[self.item_id]
            SimpleEditorBasePage.openEditor(self)
            self._editor.showEditor(
                item=self.selectedItem, item_id=self.item_id)

    def checkData(self):
        if len(self.listOfItems) > 0:
            self._valid = True
            self.emit(QtCore.SIGNAL('completeChanged()'))
        else:
            self._valid = False
            self.emit(QtCore.SIGNAL('completeChanged()'))

    def isComplete(self):
        return self._valid

    def initializePage(self):
        SimpleEditorBasePage.initializePage(self)
        self.wizard().__setitem__("poolList", self._getItemList)
        if len(self.listOfItems) == 0:
            sardana = self.wizard()["sardana"]
            _pool = Pool()
            _pool.setInstanceName(sardana)
            _pool.setLevel("1")
            _pool.setPoolDeviceName("pool/" + sardana + "/1")
            _pool.setAlias("Pool_" + sardana)
            _pool.setPoolVersion("1.2.0")
            #_pool.setPoolPath(SardanaManager.get_default_pool_path())      ###
            self.addItem(_pool.copy())


class PoolEditor(object):
    """
    Pool editor
    """

    def __init__(self, parent=None):

        self._parent = parent
        self._isEditorOpened = False
        self._pool = Pool()
        self.poolEditor = QtGui.QDialog()
        self.ui = pool_editor_UI.Ui_PoolEditor()
        self.poolEditor.setModal(True)
        self.ui.setupUi(self.poolEditor)
        self._regExp = QtCore.QRegExp("^[0-9a-zA-Z_/]{,50}")
        self._regValid = QtGui.QRegExpValidator(self._regExp, self.poolEditor)
        self._regExp2 = QtCore.QRegExp("^[0-9.]{,50}")
        self._versionValid = QtGui.QRegExpValidator(
            self._regExp2, self.poolEditor)
        self._regExp3 = QtCore.QRegExp("^[0-9a-zA-Z_/]{,100}")
        self._deviceNameValid = QtGui.QRegExpValidator(
            self._regExp3, self.poolEditor)
        self.ui.instanceNameLineEdit.setValidator(self._regValid)
        self.ui.poolDeviceNameLineEdit.setValidator(self._deviceNameValid)
        self.ui.aliasLineEdit.setValidator(self._regValid)
        self.ui.poolVersionLineEdit.setValidator(self._versionValid)
        self._item_id = None
        self._path_id = None
        self.ui.addButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-add"))
        self.ui.removeButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-remove"))
        self.ui.upButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("go-up"))
        self.ui.downButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("go-down"))
        # connections
        QtCore.QObject.connect(self.ui.createButton,
                               QtCore.SIGNAL("clicked()"), self._createPool)
        QtCore.QObject.connect(
            self.ui.addButton, QtCore.SIGNAL("clicked()"), self._addPath)
        QtCore.QObject.connect(self.ui.removeButton,
                               QtCore.SIGNAL("clicked()"), self._removePath)
        QtCore.QObject.connect(
            self.ui.upButton, QtCore.SIGNAL("clicked()"), self._moveUp)
        QtCore.QObject.connect(
            self.ui.downButton, QtCore.SIGNAL("clicked()"), self._moveDown)
        QtCore.QObject.connect(self.ui.instanceNameLineEdit, QtCore.SIGNAL(
            'textEdited(const QString &)'), self._letterChanged)
        QtCore.QObject.connect(self.ui.poolDeviceNameCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.aliasCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.poolVersionCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.poolPathList, QtCore.SIGNAL(
            "itemDoubleClicked(QListWidgetItem*)"), self._editPath)

    def _letterChanged(self):
        if self.ui.poolDeviceNameCheckBox.isChecked():
            self.ui.poolDeviceNameLineEdit.setText(
                "pool/" + self.ui.instanceNameLineEdit.text() + "/1")
        if self.ui.aliasCheckBox.isChecked():
            self.ui.aliasLineEdit.setText(
                "Pool_" + self.ui.instanceNameLineEdit.text())
        if self.ui.poolVersionCheckBox.isChecked():
            self.ui.poolVersionLineEdit.setText("0.3.0")

    def _addPath(self):
        text, ok = QtGui.QInputDialog.getText(
            self.poolEditor, 'Input Dialog', 'Type directory to be added:')
        if (ok and len(text) > 0):
            self._pool.getPoolPath().append(str(text))
            self._refreshPathList()

    def _editPath(self):
        if len(self.ui.poolPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.poolPathList.selectedIndexes()[0].row()
            text, ok = QtGui.QInputDialog.getText(
                self.poolEditor, 'Input Dialog', 'Edit selected directory:', QtGui.QLineEdit.Normal, self._pool.getPoolPath()[self._path_id])
            if (ok and len(text) > 0):
                self._pool.getPoolPath()[self._path_id] = text
        self._refreshPathList()

    def _removePath(self):
        if len(self.ui.poolPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.poolPathList.selectedIndexes()[0].row()
            self._pool.getPoolPath().remove(
                self._pool.getPoolPath()[self._path_id])
            self._refreshPathList()

    def _moveUp(self):
        if len(self.ui.poolPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.poolPathList.selectedIndexes()[0].row()
            if self._path_id > 0:
                tmp = self._pool.getPoolPath()[self._path_id]
                self._pool.getPoolPath()[self._path_id] = self._pool.getPoolPath()[
                    self._path_id - 1]
                self._pool.getPoolPath()[self._path_id - 1] = tmp
                self._refreshPathList()
                self.ui.poolPathList.setCurrentIndex(self.ui.poolPathList.indexFromItem(
                    self.ui.poolPathList.item(self._path_id - 1)))

    def _moveDown(self):
        if len(self.ui.poolPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.poolPathList.selectedIndexes()[0].row()
            if self._path_id < self.ui.poolPathList.count() - 1:
                tmp = self._pool.getPoolPath()[self._path_id]
                self._pool.getPoolPath()[self._path_id] = self._pool.getPoolPath()[
                    self._path_id + 1]
                self._pool.getPoolPath()[self._path_id + 1] = tmp
                self._refreshPathList()
                self.ui.poolPathList.setCurrentIndex(self.ui.poolPathList.indexFromItem(
                    self.ui.poolPathList.item(self._path_id + 1)))

    def _refreshPathList(self):
        self.ui.poolPathList.clear()
        for path in self._pool.getPoolPath():
            self.ui.poolPathList.addItem(path)

    def _validate(self):
        err = 0
        warnMess = ""

        if self.ui.instanceNameLineEdit.text() == "":
            err = 1
            warnMess += "Please type the Instance Name\n"
        else:
            if self._parent.checkName(self.ui.instanceNameLineEdit.text(), id=self._item_id):
                pass
            else:
                err = 1
                warnMess += "The Instance Name already exist in the DataBase\n"

        pdn = str(self.ui.poolDeviceNameLineEdit.text())
        allowed = re.compile("^.{1,}/.{1,}/.{1,}$")
        if not ((len(pdn) > 0) and (allowed.match(pdn)) and (pdn.count("/") == 2)):
            err = 1
            warnMess += "The Pool Device Name is not valid\n"

        # if self.ui.aliasLineEdit.text() == "":
        #    err=1
        #    warnMess+="Please type the Alias\n"

        if self.ui.poolVersionLineEdit.text() == "":
            err = 1
            warnMess += "Please type the Pool Version\n"

        if err == 1:
            msgBox = QtGui.QMessageBox()
            msgBox.setText("Warning !!!")
            msgBox.setStandardButtons(QtGui.QMessageBox().Ok)
            msgBox.setInformativeText(warnMess)
            msgBox.setIcon(QtGui.QMessageBox.Warning)
            ret = msgBox.exec_()
            return False
        else:
            self._pool.setInstanceName(self.ui.instanceNameLineEdit.text())
            self._pool.setPoolDeviceName(self.ui.poolDeviceNameLineEdit.text())
            self._pool.setAlias(self.ui.aliasLineEdit.text())
            self._pool.setPoolVersion(self.ui.poolVersionLineEdit.text())
            return True

    def _createPool(self):
        if self._item_id is None:
            if self._validate():
                self._parent.addItem(self._pool.copy())
        else:
            if self._validate():
                self._parent.editItem(item=self._pool.copy(), id=self._item_id)

    def showEditor(self, item=None, item_id=None):
        self._pool = Pool()
        if item is None:
            self._pool.setPoolPath(SardanaManager.get_default_pool_path())
            self._refreshPathList()
            self.poolEditor.setWindowTitle("Create New Pool")
            self.ui.instanceNameLineEdit.setText("")
            self.ui.poolDeviceNameLineEdit.setText("")
            self.ui.poolDeviceNameLineEdit.setEnabled(False)
            self.ui.poolDeviceNameCheckBox.setChecked(True)
            self.ui.aliasLineEdit.setText("")
            self.ui.aliasLineEdit.setEnabled(False)
            self.ui.aliasCheckBox.setChecked(True)
            self.ui.poolVersionLineEdit.setText("")
            self.ui.poolVersionLineEdit.setEnabled(False)
            self.ui.poolVersionCheckBox.setChecked(True)
            self.ui.createButton.setText("Create")
            self._item_id = None
            self.poolEditor.setModal(True)
            self.poolEditor.exec_()
        else:
            self._pool.setPoolPath(item.getPoolPath()[:])
            self._pool.setHostName(item.getHostName())
            self._pool.setLevel(item.getLevel())
            self._refreshPathList()
            self.poolEditor.setWindowTitle("Edit Pool")
            self.ui.instanceNameLineEdit.setText(item.getInstanceName())
            self.ui.poolDeviceNameLineEdit.setText(item.getPoolDeviceName())
            self.ui.poolDeviceNameLineEdit.setEnabled(True)
            self.ui.poolDeviceNameCheckBox.setChecked(False)
            self.ui.aliasLineEdit.setText(item.getAlias())
            self.ui.aliasLineEdit.setEnabled(True)
            self.ui.aliasCheckBox.setChecked(False)
            self.ui.poolVersionLineEdit.setText(item.getPoolVersion())
            self.ui.poolVersionLineEdit.setEnabled(True)
            self.ui.poolVersionCheckBox.setChecked(True)
            self.ui.createButton.setText("Edit")
            self._item_id = item_id
            self.poolEditor.setModal(True)
            self.poolEditor.exec_()


class MS(Item):
    """
    Represents a Macroserver
    """

    def __init__(self, instanceName="", hostName="", level=None, poolName="", msDeviceName="", msAlias="", msVersion="", msPath=[], doorName="", doorAlias=""):
        Item.__init__(self, instanceName, hostName, level)
        self._msDeviceName = str(msDeviceName)
        self._msAlias = str(msAlias)
        self._msVersion = str(msVersion)
        self._msPath = map(str, msPath)
        self._doorName = str(doorName)
        self._doorAlias = str(doorAlias)
        self._poolName = str(poolName)

    def setPoolName(self, name):
        self._poolName = str(name)

    def getPoolName(self):
        return self._poolName

    def setMSDeviceName(self, name):
        self._msDeviceName = str(name)

    def getMSDeviceName(self):
        return self._msDeviceName

    def setMSAlias(self, alias):
        self._msAlias = str(alias)

    def getMSAlias(self):
        return self._msAlias

    def setMSVersion(self, msVersion):
        self._msVersion = str(msVersion)

    def getMSVersion(self):
        return self._msVersion

    def setMSPath(self, path):
        self._msPath = map(str, path)

    def getMSPath(self):
        return self._msPath

    def setDoorName(self, doorName):
        self._doorName = str(doorName)

    def getDoorName(self):
        return self._doorName

    def setDoorAlias(self, doorAlias):
        self._doorAlias = str(doorAlias)

    def getDoorAlias(self):
        return self._doorAlias

    def text(self):
        return str(self._instanceName)

    def copy(self):
        return MS(self._instanceName, self._hostName, self._level, self._poolName, self._msDeviceName, self._msAlias, self._msVersion, self._msPath[:], self._doorName, self._doorAlias)


class MSEditor(object):
    """
    Macroserver editor
    """

    def __init__(self, parent=None):
        self._parent = parent
        #self._parent = weakref.ref(parent)
        self._isEditorOpened = False
        self._ms = MS()
        self.msEditor = QtGui.QDialog()
        self.ui = ms_editor_UI.Ui_MSEditor()
        self.msEditor.setModal(True)
        self.ui.setupUi(self.msEditor)
        self._regExp = QtCore.QRegExp("^[0-9a-zA-Z]{,50}")
        self._regValid = QtGui.QRegExpValidator(self._regExp, self.msEditor)
        self._regExp2 = QtCore.QRegExp("^[0-9.]{,50}")
        self._versionValid = QtGui.QRegExpValidator(
            self._regExp2, self.msEditor)
        self._regExp3 = QtCore.QRegExp("^[0-9a-zA-Z_/]{,100}")
        self._deviceNameValid = QtGui.QRegExpValidator(
            self._regExp3, self.msEditor)
        self.ui.instanceNameLineEdit.setValidator(self._regValid)
        self.ui.msDeviceNameLineEdit.setValidator(self._deviceNameValid)
        self.ui.msAliasLineEdit.setValidator(self._regValid)
        self.ui.msVersionLineEdit.setValidator(self._versionValid)
        self.ui.doorNameLineEdit.setValidator(self._regValid)
        self.ui.doorAliasLineEdit.setValidator(self._regValid)
        self._item_id = None
        self._path_id = None
        self.ui.addButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-add"))
        self.ui.removeButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-remove"))
        self.ui.upButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("go-up"))
        self.ui.downButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("go-down"))
        # connections
        QtCore.QObject.connect(self.ui.createButton,
                               QtCore.SIGNAL("clicked()"), self._createMS)
        QtCore.QObject.connect(
            self.ui.addButton, QtCore.SIGNAL("clicked()"), self._addPath)
        QtCore.QObject.connect(self.ui.removeButton,
                               QtCore.SIGNAL("clicked()"), self._removePath)
        QtCore.QObject.connect(
            self.ui.upButton, QtCore.SIGNAL("clicked()"), self._moveUp)
        QtCore.QObject.connect(
            self.ui.downButton, QtCore.SIGNAL("clicked()"), self._moveDown)
        QtCore.QObject.connect(self.ui.instanceNameLineEdit, QtCore.SIGNAL(
            'textEdited(const QString &)'), self._letterChanged)
        QtCore.QObject.connect(self.ui.msDeviceNameCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.msAliasCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.msVersionCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.doorNameCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.doorAliasCheckBox, QtCore.SIGNAL(
            "toggled(bool)"), self._letterChanged)
        QtCore.QObject.connect(self.ui.msPathList, QtCore.SIGNAL(
            "itemDoubleClicked(QListWidgetItem*)"), self._editPath)

    def _letterChanged(self):
        if self.ui.msDeviceNameCheckBox.isChecked():
            self.ui.msDeviceNameLineEdit.setText(
                "macroserver/" + self.ui.instanceNameLineEdit.text() + "/1")
        if self.ui.msAliasCheckBox.isChecked():
            self.ui.msAliasLineEdit.setText(
                "MS_" + self.ui.instanceNameLineEdit.text())
        if self.ui.msVersionCheckBox.isChecked():
            self.ui.msVersionLineEdit.setText("0.3.0")
        if self.ui.doorNameCheckBox.isChecked():
            self.ui.doorNameLineEdit.setText(
                "door/" + self.ui.instanceNameLineEdit.text() + "/1")
        if self.ui.doorAliasCheckBox.isChecked():
            self.ui.doorAliasLineEdit.setText(
                "Door_" + self.ui.instanceNameLineEdit.text())

    def getPoolServerList(self):
        return self._parent.getPoolServerList()

    def fillPoolNameCB(self, selected=None):
        self.ui.poolNameComboBox.clear()
        i = 0
        if selected is not None:
            if selected == "":
                selected = "None"
            self.ui.poolNameComboBox.addItem(selected)
            i += 1
        for item in self._parent.wizard()["poolList"]:
            self.ui.poolNameComboBox.addItem(item.text())
            i += 1
        self.ui.poolNameComboBox.insertSeparator(i)
        i += 1
        for item in self.getPoolServerList():
            self.ui.poolNameComboBox.addItem(item)
            i += 1
        self.ui.poolNameComboBox.insertSeparator(i)
        self.ui.poolNameComboBox.addItem("None")

    def db(self):
        return self._parent.wizard()['db']

    def _editPath(self):
        if len(self.ui.msPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.msPathList.selectedIndexes()[0].row()
            text, ok = QtGui.QInputDialog.getText(
                self.msEditor, 'Input Dialog', 'Edit selected directory:', QtGui.QLineEdit.Normal, self._ms.getMSPath()[self._path_id])
            if (ok and len(text) > 0):
                self._ms.getMSPath()[self._path_id] = text
        self._refreshPathList()

    def _addPath(self):
        text, ok = QtGui.QInputDialog.getText(
            self.msEditor, 'Input Dialog', 'Type directory to be added:')
        if (ok and len(text) > 0):
            self._ms.getMSPath().append(str(text))
            self._refreshPathList()

    def _removePath(self):
        if len(self.ui.msPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.msPathList.selectedIndexes()[0].row()
            self._ms.getMSPath().remove(self._ms.getMSPath()[self._path_id])
            self._refreshPathList()

    def _moveUp(self):
        if len(self.ui.msPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.msPathList.selectedIndexes()[0].row()
            if self._path_id > 0:
                tmp = self._ms.getMSPath()[self._path_id]
                self._ms.getMSPath()[self._path_id] = self._ms.getMSPath()[
                    self._path_id - 1]
                self._ms.getMSPath()[self._path_id - 1] = tmp
                self._refreshPathList()
                self.ui.msPathList.setCurrentIndex(self.ui.msPathList.indexFromItem(
                    self.ui.msPathList.item(self._path_id - 1)))

    def _moveDown(self):
        if len(self.ui.msPathList.selectedIndexes()) > 0:
            self._path_id = self.ui.msPathList.selectedIndexes()[0].row()
            if self._path_id < self.ui.msPathList.count() - 1:
                tmp = self._ms.getMSPath()[self._path_id]
                self._ms.getMSPath()[self._path_id] = self._ms.getMSPath()[
                    self._path_id + 1]
                self._ms.getMSPath()[self._path_id + 1] = tmp
                self._refreshPathList()
                self.ui.msPathList.setCurrentIndex(self.ui.msPathList.indexFromItem(
                    self.ui.msPathList.item(self._path_id + 1)))

    def _refreshPathList(self):
        self.ui.msPathList.clear()
        for path in self._ms.getMSPath():
            self.ui.msPathList.addItem(path)

    def _validate(self):
        err = 0
        warnMess = ""

        if self.ui.instanceNameLineEdit.text() == "":
            err = 1
            warnMess += "Please type the Instance Name\n"
        else:
            if self._parent.checkName(name=self.ui.instanceNameLineEdit.text(), id=self._item_id):
                pass
            else:
                err = 1
                warnMess += "The Instance Name already exist in the DataBase\n"

        if self.ui.msDeviceNameLineEdit.text() == "":
            err = 1
            warnMess += "Please type the Macro Server Device Name\n"

        mdn = str(self.ui.msDeviceNameLineEdit.text())
        allowed = re.compile("^.{1,}/.{1,}/.{1,}$")
        if not ((len(mdn) > 0) and (allowed.match(mdn)) and (mdn.count("/") == 2)):
            err = 1
            warnMess += "The Macro Server Device Name is not valid\n"

        # if self.ui.msAliasLineEdit.text() == "":
        #    err=1
        #    warnMess+="Please type the Macro Server Alias\n"

        if self.ui.msVersionLineEdit.text() == "":
            err = 1
            warnMess += "Please type the Macro Server Version\n"

        dn = str(self.ui.doorNameLineEdit.text())
        allowed = re.compile("^.{1,}/.{1,}/.{1,}$")
        if not ((len(dn) > 0) and (allowed.match(dn)) and (dn.count("/") == 2)):
            err = 1
            warnMess += "The Door Name is not valid\n"

        # if self.ui.doorAliasLineEdit.text() == "":
        #    err=1
        #    warnMess+="Please type the Door Alias\n"

        if err == 1:
            msgBox = QtGui.QMessageBox()
            msgBox.setText("Warning !!!")
            msgBox.setStandardButtons(QtGui.QMessageBox().Ok)
            msgBox.setInformativeText(warnMess)
            msgBox.setIcon(QtGui.QMessageBox.Warning)
            ret = msgBox.exec_()
            return False
        else:
            self._ms.setInstanceName(self.ui.instanceNameLineEdit.text())
            self._ms.setMSDeviceName(self.ui.msDeviceNameLineEdit.text())
            self._ms.setMSAlias(self.ui.msAliasLineEdit.text())
            self._ms.setMSVersion(self.ui.msVersionLineEdit.text())
            self._ms.setDoorName(self.ui.doorNameLineEdit.text())
            self._ms.setDoorAlias(self.ui.doorAliasLineEdit.text())
            if self.ui.poolNameComboBox.currentText() == "None":
                self._ms.setPoolName("")
            else:
                self._ms.setPoolName(self.ui.poolNameComboBox.currentText())

            return True

    def _createMS(self):
        if self._item_id is None:
            if self._validate():
                self._parent.addItem(self._ms.copy())
        else:
            if self._validate():
                self._parent.editItem(item=self._ms.copy(), id=self._item_id)

    def showEditor(self, item=None, item_id=None):
        self._ms = MS()
        if item is None:
            self._ms.setMSPath(SardanaManager.get_default_ms_path())
            self._refreshPathList()
            self.msEditor.setWindowTitle("Create New Macro Server")
            self.ui.instanceNameLineEdit.setText("")
            self.fillPoolNameCB()
            self.ui.msDeviceNameLineEdit.setText("")
            self.ui.msDeviceNameLineEdit.setEnabled(False)
            self.ui.msDeviceNameCheckBox.setChecked(True)
            self.ui.msAliasLineEdit.setText("")
            self.ui.msAliasLineEdit.setEnabled(False)
            self.ui.msAliasCheckBox.setChecked(True)
            self.ui.msVersionLineEdit.setText("")
            self.ui.msVersionLineEdit.setEnabled(False)
            self.ui.msVersionCheckBox.setChecked(True)
            self.ui.doorNameLineEdit.setText("")
            self.ui.doorNameLineEdit.setEnabled(False)
            self.ui.doorNameCheckBox.setChecked(True)
            self.ui.doorAliasLineEdit.setText("")
            self.ui.doorAliasLineEdit.setEnabled(False)
            self.ui.doorAliasCheckBox.setChecked(True)
            self.ui.createButton.setText("Create")
            self._item_id = None
            self.msEditor.setModal(True)
            self.msEditor.exec_()
        else:
            self._ms.setMSPath(item.getMSPath()[:])
            self._ms.setHostName(item.getHostName())
            self._ms.setLevel(item.getLevel())
            self._refreshPathList()
            self.msEditor.setWindowTitle("Edit Macro Server")
            self.ui.instanceNameLineEdit.setText(item.getInstanceName())
            self.fillPoolNameCB(selected=item.getPoolName())
            self.ui.msDeviceNameLineEdit.setText(item.getMSDeviceName())
            self.ui.msDeviceNameLineEdit.setEnabled(True)
            self.ui.msDeviceNameCheckBox.setChecked(False)
            self.ui.msAliasLineEdit.setText(item.getMSAlias())
            self.ui.msAliasLineEdit.setEnabled(True)
            self.ui.msAliasCheckBox.setChecked(False)
            self.ui.msVersionLineEdit.setText(item.getMSVersion())
            self.ui.msVersionLineEdit.setEnabled(True)
            self.ui.msVersionCheckBox.setChecked(True)
            self.ui.doorNameLineEdit.setText(item.getDoorName())
            self.ui.doorNameLineEdit.setEnabled(True)
            self.ui.doorNameCheckBox.setChecked(False)
            self.ui.doorAliasLineEdit.setText(item.getDoorAlias())
            self.ui.doorAliasLineEdit.setEnabled(True)
            self.ui.doorAliasCheckBox.setChecked(False)
            self.ui.createButton.setText("Edit")
            self._item_id = item_id
            self.msEditor.setModal(True)
            self.msEditor.exec_()


class AddMSBasePage(SimpleEditorBasePage):
    """
    Page for editing the list of Macroservers
    """

    def __init__(self, parent=None):
        self._editor = None
        SimpleEditorBasePage.__init__(self, parent)
        self._editor = MSEditor(parent=self)
        self.selectedItem = None
        self.item_id = None
        self.setSubTitle(
            'You can use this manager if you would like to Add, Edit or Delete Macro Server entries in the database.')
        self.addButton.setText(QtGui.QApplication.translate(
            "Form", "Add MS", None, QtGui.QApplication.UnicodeUTF8))
        self.editButton.setText(QtGui.QApplication.translate(
            "Form", "Edit MS", None, QtGui.QApplication.UnicodeUTF8))
        self.removeButton.setText(QtGui.QApplication.translate(
            "Form", "Remove MS", None, QtGui.QApplication.UnicodeUTF8))
        # connections
        QtCore.QObject.connect(self.addButton, QtCore.SIGNAL(
            "clicked()"), self.openAddEditor)
        QtCore.QObject.connect(self.editButton, QtCore.SIGNAL(
            "clicked()"), self.openItemEditor)
        QtCore.QObject.connect(
            self.removeButton, QtCore.SIGNAL("clicked()"), self.delete)
        QtCore.QObject.connect(self.tableWidget, QtCore.SIGNAL(
            "cellDoubleClicked(int,int)"), self.openItemEditor)

    def delete(self):
        if len(self.tableWidget.selectedIndexes()) > 0:
            msgBox = QtGui.QMessageBox()
            msgBox.setText("The list of Macro Servers has been modified.")
            self.item_id = self.tableWidget.selectedIndexes()[0].row()
            self.selectedItem = self.listOfItems[self.item_id]
            msgBox.setInformativeText(
                "Do you want to delete Macro Server?:\n" + self.selectedItem.text())
            msgBox.setStandardButtons(
                QtGui.QMessageBox().Ok | QtGui.QMessageBox().Cancel)
            msgBox.setDefaultButton(QtGui.QMessageBox().Cancel)
            msgBox.setIcon(QtGui.QMessageBox.Question)
            ret = msgBox.exec_()
            if ret == QtGui.QMessageBox().Ok:
                self.removeItem(self.item_id)
            if ret == QtGui.QMessageBox().Cancel:
                pass

    def openAddEditor(self):
        self.item_id = None
        self.selectedItem = None
        SimpleEditorBasePage.openEditor(self)
        self._editor.showEditor(item=None, item_id=None)

    def openItemEditor(self):
        if len(self.tableWidget.selectedIndexes()) > 0:
            self.item_id = self.tableWidget.selectedIndexes()[0].row()
            self.selectedItem = self.listOfItems[self.item_id]
            SimpleEditorBasePage.openEditor(self)
            self._editor.showEditor(
                item=self.selectedItem, item_id=self.item_id)

    def checkData(self):
        pass

    def isComplete(self):
        return self._valid

    def _fillList(self):
        warnMess = ""
        err = False

        for item in self.listOfItems:
            if len(item.getPoolName()) > 0:
                poolExist = False
                for sv_item in self.getPoolServerList():
                    if item.getPoolName() == sv_item:
                        poolExist = True
                        break
                if not poolExist:
                    for cr_item in self.wizard()["poolList"]:
                        if item.getPoolName() == cr_item.getInstanceName():
                            poolExist = True
                            break
                if not poolExist:
                    warnMess += "Pool: '" + item.getPoolName() + "' does not exist \n"
                    item.setPoolName("")
                    err = True

        if err:
            msgBox = QtGui.QMessageBox()
            msgBox.setText("Warning !!!")
            msgBox.setStandardButtons(QtGui.QMessageBox().Ok)
            msgBox.setInformativeText(warnMess)
            msgBox.setIcon(QtGui.QMessageBox.Warning)
            ret = msgBox.exec_()

        SimpleEditorBasePage._fillList(self)

    def initializePage(self):
        SimpleEditorBasePage.initializePage(self)
        self.wizard().__setitem__("msList", self._getItemList)
        if len(self.listOfItems) == 0:
            sardana = self.wizard()["sardana"]
            _ms = MS()
            _ms.setInstanceName(sardana)
            _ms.setPoolName(self.wizard()["poolList"][0].getInstanceName())
            _ms.setLevel("1")
            _ms.setMSDeviceName("macroserver/" + sardana + "/1")
            _ms.setMSAlias("MS_" + sardana)
            _ms.setMSVersion("0.3.0")
            _ms.setDoorName("door/" + sardana + "/1")
            _ms.setDoorAlias("Door_" + sardana)
            #_ms.setMSPath(SardanaManager.get_default_ms_path())            ###

            self.addItem(_ms.copy())
        self._editor.fillPoolNameCB()


class SardanaCommitBasePage(wiz.SardanaIntroBasePage):
    """
    Commiting page
    """

    def __init__(self, parent=None):
        QtGui.QWizardPage.__init__(self, parent)
        self.setPixmap(QtGui.QWizard.WatermarkPixmap,
                       QtGui.QPixmap(":/watermark.jpg"))
        self._layout = QtGui.QFormLayout()
        self._sardanaNameLabel = QtGui.QLabel()
        self._hostNameLabel = QtGui.QLabel()
        self._portLabel = QtGui.QLabel()
        self._poollsLabel = QtGui.QLabel()
        self._mssLabel = QtGui.QLabel()
        self._layout.addRow(self._set_style(
            QtGui.QLabel("Sardana:")), self._sardanaNameLabel)
        self._layout.addRow(self._set_style(
            QtGui.QLabel("Host:")), self._hostNameLabel)
        self._layout.addRow(self._set_style(
            QtGui.QLabel("Port:")), self._portLabel)
        self._layout.addRow(self._set_style(
            QtGui.QLabel("")), QtGui.QLabel(""))
        self._layout.addRow(self._set_style(
            QtGui.QLabel("Pools:")), self._poollsLabel)
        self._layout.addRow(self._set_style(
            QtGui.QLabel("MacroServers:")), self._mssLabel)
        self._checkBox = QtGui.QCheckBox(
            "Start local server automatically\neven if no starter is present")
        self._checkBox.setChecked(True)
        self._layout.addRow(self._checkBox)
        self._poolListText = ""
        self._msListText = ""
        self.setLayout(self._layout)
        # self.setCommitPage(True)
        self.setTitle('Confirmation.')

    def next(self):
        QWizard.next(self)

    def _set_style(self, w):
        f = w.font()
        f.setBold(True)
        w.setFont(f)
        return w

    def _getAutoStart(self):
        return self._checkBox.checkState()

    def initializePage(self):
        wiz.SardanaIntroBasePage.initializePage(self)
        self.wizard().__setitem__("autoStart", self._getAutoStart)
        self._checkBox.setVisible(False)

        for item in self.wizard()["poolList"] + self.wizard()["msList"]:
            if item.getHostName() == "localhost":
                if SardanaManager().has_localhost_starter():
                    self._checkBox.setVisible(True)
                    break

        self._sardanaNameLabel.setText(self.wizard()["sardana"])
        self._hostNameLabel.setText(self.wizard()["host"])
        self._portLabel.setText(self.wizard()["port"])
        self._poolListText = ""
        # limit the length of the list
        if len(self.wizard()["poolList"]) + len(self.wizard()["msList"]) > 14:
            limited = True
        else:
            limited = False
        i = 0
        for pool in self.wizard()["poolList"]:
            if (limited) and (i > 6):
                self._poolListText += "...\n"
                break
            self._poolListText += pool.getInstanceName() + "\n"
            i += 1

        self._poollsLabel.setText(self._poolListText)
        self._msListText = ""
        i = 0
        for ms in self.wizard()["msList"]:
            if (limited) and (i > 6):
                self._msListText += "...\n"
                break
            self._msListText += ms.getInstanceName() + "\n"
            i += 1

        self._mssLabel.setText(self._msListText)

    def setNextPageId(self, id):
        self._nextPageId = id


class SardanaOutroBasePage(wiz.SardanaBasePage):
    """
    The last page in the wizard for creation sardana and all of the pools and macroservers that it consist
    """

    def __init__(self, parent=None):
        QtGui.QWizardPage.__init__(self, parent)
        self.setPixmap(QtGui.QWizard.WatermarkPixmap,
                       QtGui.QPixmap(":/watermark.jpg"))
        self._valid = True
        self._layout = QtGui.QVBoxLayout()
        self._spacerItem = QtGui.QSpacerItem(
            20, 40, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self._layout.addItem(self._spacerItem)
        self._horizontalLayout_1 = QtGui.QHBoxLayout()
        self._label = QtGui.QLabel("\nInitializing...")
        self._label.setWordWrap(True)
        self._horizontalLayout_1.addWidget(self._label)
        self._layout.addLayout(self._horizontalLayout_1)
        self._pbar = QtGui.QProgressBar(self)
        self._layout.addWidget(self._pbar)
        self._horizontalLayout_2 = QtGui.QHBoxLayout()
        self._spacerItem1 = QtGui.QSpacerItem(
            40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self._horizontalLayout_2.addItem(self._spacerItem1)
        self._cancelButton = QtGui.QPushButton("Cancel")
        self._horizontalLayout_2.addWidget(self._cancelButton)
        self._spacerItem2 = QtGui.QSpacerItem(
            40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self._horizontalLayout_2.addItem(self._spacerItem2)
        self._layout.addLayout(self._horizontalLayout_2)
        self._spacerItem3 = QtGui.QSpacerItem(
            20, 40, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self._layout.addItem(self._spacerItem3)
        self.setLayout(self._layout)
#        self.setTitle('The new instance of Sardana has been successfully created.')

    def initializePage(self):
        self.wizard().setOption(QtGui.QWizard.NoCancelButton, True)
        self.wizard().setOption(QtGui.QWizard.NoBackButtonOnLastPage, True)
        self.progress = Progress(self.wizard()["db"], self.wizard()["sardana"], self.wizard()[
                                 "poolList"][:], self.wizard()["msList"][:], self.wizard()["autoStart"])
        Qt.QObject.connect(self.progress, Qt.SIGNAL(
            "valueUpdated"), self.setProgress)
        QtCore.QObject.connect(self._cancelButton, QtCore.SIGNAL(
            "clicked()"), self.progress.cancel)
        self.progress.start()
        self._valid = False
        self._cancelButton.show()

    def isComplete(self):
        return self._valid

    def setProgress(self, message, value, status):
        # status
        #  0 - normal
        #  1 - finished
        # -1 - interrupted
        if status == 0:
            self.setTitle(
                'The Wizard is creating the new instance of Sardana.')
        if status == 1:
            self.setTitle(
                'The new instance of Sardana has been successfully created.')
            self._cancelButton.hide()
            self._valid = True
            self.emit(QtCore.SIGNAL('completeChanged()'))
        if status == -1:
            self.setTitle('The new instance of Sardana has NOT been created.')
            self.wizard().setOption(QtGui.QWizard.NoBackButtonOnLastPage, False)
            self.wizard().setOption(QtGui.QWizard.NoCancelButton, False)
            self._cancelButton.hide()
            self._valid = False
            self.emit(QtCore.SIGNAL('completeChanged()'))

        self._label.setText(message)
        self._pbar.setValue(value)


class Progress (Qt.QThread):
    """
    Indicate the percent of sardana creation process
    """

    def __init__(self, db, sardana, poolList, msList, autoStart):
        Qt.QThread.__init__(self)
        self._sardana = sardana
        self._poolList = poolList
        self._poolNames = [x.getPoolDeviceName() for x in poolList]
        self._msList = msList
        self._cancelled = False
        self._autoStart = autoStart
        self._db = db

    def cancel(self):
        self._cancelled = True

    def run(self):
        # status
        #  0 - normal
        #  1 - finished
        # -1 - interrupted
        for msg, percentage, status in self._createSardana():
            self.emit(Qt.SIGNAL("valueUpdated"), msg, percentage, status)

    def _createSardana(self):
        err = False
        mess = "Click the Finish button to exit this wizard."
        sardanaManager = SardanaManager()
        self._newSardana = None

        percent = 0
        step = 100 / (len(self._poolList) + len(self._msList) + 1)

        if len(self._msList) > 0:
            try:
                self._sardana_device_name = str(
                    self._msList[0].getMSDeviceName())
                self._newSardana = sardanaManager.create_sardana(
                    str(self._sardana), self._sardana_device_name, db=self._db)
            except Exception, e:
                err = True
                mess = str(e)
                traceback.print_exc()
        else:
            try:
                self._sardana_device_name = str(
                    self._poolList[0].getPoolDeviceName())
                self._newSardana = sardanaManager.create_sardana(
                    str(self._sardana), self._sardana_device_name, db=self._db)
            except Exception, e:
                err = True
                mess = str(e)

        if err:
            #msgbox = TaurusMessageBox(*sys.exc_info())
            # msgbox.exec_()

            yield "\nCancelled...", 100, 1

        else:  # Sardana has been created
            percent += step
            yield "\nProcessing the list of Pools", int(percent), 0
            time.sleep(0.5)

            try:

                for pool in self._poolList:
                    if self._cancelled:
                        raise Exception("Canceled by user")

                    yield "Adding Pool:\n %s" % pool.getInstanceName(), int(percent), 0

                    if pool.getPoolDeviceName() == "":
                        pool.setPoolDeviceName(None)
                    if pool.getAlias() == "":
                        pool.setAlias(None)
                    newPool = self._newSardana.create_pool(pool.getInstanceName(), pool.getPoolPath(
                    ), pool.getPoolVersion(), alias=pool.getAlias(), device_name=pool.getPoolDeviceName())
                    if pool.getHostName() not in (None, '', 'None'):
                        if pool.getHostName() == "localhost":
                            if self._autoStart:
                                yield "Starting Pool:\n %s" % pool.getInstanceName(), int(percent), 0
                                newPool.local_run()
                        else:
                            yield "Starting Pool:\n %s" % pool.getInstanceName(), int(percent), 0
                            newPool.starter_run(
                                pool.getHostName(), pool.getLevel())
                    percent += step

                yield "\nProcessing the list of MacroServers", int(percent), 0
                time.sleep(0.5)

                for ms in self._msList:
                    if self._cancelled:
                        raise Exception("Canceled by user")

                    yield "Adding MacroServer:\n %s" % ms.getInstanceName(), int(percent), 0

                    if ms.getMSDeviceName() == "":
                        ms.setMSDeviceName(None)
                    if ms.getMSAlias() == "":
                        ms.setMSAlias(None)
                    newMS = self._newSardana.create_macroserver(ms.getInstanceName(), ms.getMSPath(
                    ), self._poolNames, ms.getMSVersion(), alias=ms.getMSAlias(), device_name=ms.getMSDeviceName())
                    newMS.create_door(ms.getDoorAlias(), ms.getDoorName())
                    if ms.getHostName() not in (None, '', 'None'):
                        if ms.getHostName() == "localhost":
                            if self._autoStart:
                                yield "Starting MacroServer:\n %s" % ms.getInstanceName(), int(percent), 0
                                newMS.local_run()
                        else:
                            yield "Starting MacroServer:\n %s" % ms.getInstanceName(), int(percent), 0
                            newMS.starter_run(ms.getHostName(), ms.getLevel())

                    percent += step

                self._newSardana.set_device_name(self._sardana_device_name)
                yield "\nDone...", 100, 1

            except Exception, e:
                err = True
                mess = str(e)
                traceback.print_exc()
                # remove Sardana if something goes wrong
                sardanaManager.remove_sardana(str(self._sardana))
                yield "\nCanceled by user", 0, -1


def addSardana():

    app = QtGui.QApplication([])
    QtCore.QResource.registerResource(wiz.get_resources())

    Pages = Enumeration('Pages', ('IntroPage', 'TangoPage', 'SardanaPage',
                                  'PoolPage', 'MSPage', 'CommitPage', 'OutroPage'))
    w = wiz.SardanaBaseWizard()

    intro = NewSardanaIntroPage()
    w.setPage(Pages.IntroPage, intro)
    intro.setNextPageId(Pages.TangoPage)

    tg_host_page = SelectTangoHostPage()
    w.setPage(Pages.TangoPage, tg_host_page)
    tg_host_page.checkData()
    tg_host_page.setNextPageId(Pages.SardanaPage)

    sardana_page = AddSardanaBasePage()
    w.setPage(Pages.SardanaPage, sardana_page)
    sardana_page.setNextPageId(Pages.PoolPage)

    pool_page = AddPoolBasePage()
    w.setPage(Pages.PoolPage, pool_page)
    pool_page.setNextPageId(Pages.MSPage)

    ms_page = AddMSBasePage()
    w.setPage(Pages.MSPage, ms_page)
    ms_page.setNextPageId(Pages.CommitPage)

    commit_page = SardanaCommitBasePage()
    w.setPage(Pages.CommitPage, commit_page)
    commit_page.setNextPageId(Pages.OutroPage)

    outro_page = SardanaOutroBasePage()
    w.setPage(Pages.OutroPage, outro_page)
    w.setOption(QtGui.QWizard.CancelButtonOnLeft, True)

    #Qt.QObject.connect(w, Qt.SIGNAL("done()"), done)
    w.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    addSardana()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
This module provides objects to manage macro-related tasks. Its primary use is
to be used within a TaurusGui for managing panels for:
- setting preferences in the sardana control system for data I/O
- displaying results of macro executions, including creating/removing panels
  for plotting results of scans
- editing macros

.. note:: This module was originally implemented in taurus as
          `taurus.qt.qtgui.taurusgui.macrolistener`
"""



from builtins import object

import datetime
import collections

import numpy

try:
    import pyqtgraph
except ImportError:
    pyqtgraph = None

from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseComponent
from taurus.core.util.containers import ArrayBuffer, LoopList

from sardana.taurus.core.tango.sardana import PlotType


__all__ = ['MacroBroker', 'DynamicPlotManager', 'assertPlotAvailability']

__docformat__ = 'restructuredtext'

COLORS = [Qt.QColor(Qt.Qt.red),
          Qt.QColor(Qt.Qt.green),
          Qt.QColor(Qt.Qt.magenta),
          Qt.QColor(Qt.Qt.blue),
          Qt.QColor(Qt.Qt.cyan),
          Qt.QColor(Qt.Qt.yellow),
          Qt.QColor(Qt.Qt.white)]

SYMBOLS = ['x', 'o', 't', 't1', 't2', 't3', 's', 'p']

CURVE_STYLES = [(color, symbol) for symbol in SYMBOLS for color in COLORS]

NO_PLOT_MESSAGE = 'Plots cannot be displayed (pyqtgraph not installed)'


def assertPlotAvailability(exit_on_error=True):
    if pyqtgraph is None:
        Qt.QMessageBox.critical(None, 'Plot error', NO_PLOT_MESSAGE)
        if exit_on_error:
            exit(1)


def empty_data(nb_points):
    return ArrayBuffer(numpy.full(nb_points, numpy.nan))


class MultiPlotWidget(Qt.QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)
        layout = Qt.QVBoxLayout(self)
        self.win = pyqtgraph.GraphicsLayoutWidget()
        layout.addWidget(self.win)
        self._plots = {}
        self._timer = None
        self._event_nb = 0
        self._last_event_nb = 0

    # plots: a list of plots
    # each plot is:
    #   dict: { x_axis: { name: axis-name, label: axis-label
    #           curves: [{ name: curve_name, label: curve_label }] }

    def prepare(self, plots, nb_points=None):
        self.win.clear()
        plot_widgets = {}
        nb_points = 2**16 if nb_points is None else nb_points
        plots_per_row = int(len(plots)**0.5)
        for idx, plot in enumerate(plots):
            plot_curves = {}
            x_axis, curves = plot['x_axis'], plot['curves']
            if idx % plots_per_row == 0:
                self.win.nextRow()
            nb_curves = len(curves)
            labels = dict(bottom=x_axis['label'])
            if nb_curves == 1:
                labels['left'] = curves[0]['label']
            plot_widget = self.win.addPlot(labels=labels)
            plot_widget.showGrid(x=True, y=True)
            if nb_curves > 1:
                plot_widget.addLegend()
            plot_widget.x_axis = dict(x_axis, data=empty_data(nb_points))
            styles = LoopList(CURVE_STYLES)
            for curve in curves:
                pen, symbol = styles.current()
                styles.next()
                style = dict(pen=pen, symbol=symbol,
                             symbolSize=5, symbolPen=pen, symbolBrush=pen)
                curve_item = plot_widget.plot(name=curve['label'], **style)
                curve_item.curve_data = empty_data(nb_points)
                plot_curves[curve['name']] = curve_item
            plot_widgets[plot_widget] = plot_curves
        self._plots = plot_widgets
        self._event_nb = 0
        self._last_event_nb = 0
        self._start_update()

    def onNewPoint(self, data):
        # update internal data and mark the new event
        for plot_widget, curves in self._plots.items():
            x_axis = plot_widget.x_axis
            x_data = x_axis['data']
            x_data.append(data[x_axis['name']])
            for curve_name, curve_item in curves.items():
                y_data = curve_item.curve_data
                y_data.append(data[curve_name])
        self._event_nb += 1

    def onEnd(self, data):
        self.do_update()
        self._end_update()

    def do_update(self):
        if self._event_nb == self._last_event_nb:
            return
        self._last_event_nb = self._event_nb
        for plot_widget, curves in self._plots.items():
            x_axis = plot_widget.x_axis
            x_data = x_axis['data']
            for curve_name, curve_item in curves.items():
                y_data = curve_item.curve_data
                curve_item.setData(x_data.contents(), y_data.contents())

    def _start_update(self):
        self._end_update()
        timer = Qt.QTimer()
        timer.timeout.connect(self.do_update)
        # refresh curves at a fix rate of ~5Hz
        timer.start(200)
        self._timer = timer

    def _end_update(self):
        if self._timer:
            self._timer.stop()
        self._timer = None


class DynamicPlotManager(Qt.QObject, TaurusBaseComponent):
    '''This is a manager of plots related to the execution of macros.
    It dynamically creates/removes plots according to the configuration made by
    an ExperimentConfiguration widget.

    Currently it supports only 1D scan trends (2D scans are only half-baked)

    To use it simply instantiate it and pass it a door name as a model. You may
    want to call :meth:`onExpConfChanged` to update the configuration being
    used.
    '''

    plots_available = pyqtgraph is not None

    newShortMessage = Qt.pyqtSignal('QString')

    Single = 'single'  # each curve has its own plot
    XAxis = 'x-axis'   # group curves with same X-Axis

    def __init__(self, parent=None):
        Qt.QObject.__init__(self, parent)
        TaurusBaseComponent.__init__(self, self.__class__.__name__)

        self._group_mode = self.XAxis
        Qt.qApp.SDM.connectWriter("shortMessage", self, 'newShortMessage')

        self._plot = MultiPlotWidget()
        self.createPanel(
            self._plot, 'Scan plot', registerconfig=False, permanent=False)

    def setGroupMode(self, group):
        assert group in (self.Single, self.XAxis)
        self._group_mode = group

    def groupMode(self):
        return self._group_mode

    def setModel(self, doorname):
        '''reimplemented from :meth:`TaurusBaseComponent`

        :param doorname: (str) device name corresponding to a Door device.
        '''
        TaurusBaseComponent.setModel(self, doorname)
        # self._onDoorChanged(doorname)
        if not doorname:
            return
        self.door = self.getModelObj()
        if not isinstance(self.door, Qt.QObject):
            msg = "Unexpected type (%s) for %s" % (repr(type(self.door)),
                                                   doorname)
            Qt.QMessageBox.critical(
                self.parent(), 'Door connection error', msg)
            return

        self._checkJsonRecorder()

        self.door.recordDataUpdated.connect(self.onRecordDataUpdated)
        self.message_template = 'Ready!'

    def _checkJsonRecorder(self):
        '''Checks if JsonRecorder env var is set and offers to set it'''
        door = self.getModelObj()
        if 'JsonRecorder' not in door.getEnvironment():
            msg = ('JsonRecorder environment variable is not set, but it '
                   + 'is needed for displaying trend plots.\n'
                   + 'Enable it globally for %s?') % door.getFullName()
            result = Qt.QMessageBox.question(
                self.parent(), 'JsonRecorder not set', msg,
                Qt.QMessageBox.Yes | Qt.QMessageBox.No)
            if result == Qt.QMessageBox.Yes:
                door.putEnvironment('JsonRecorder', True)
                self.info('JsonRecorder Enabled for %s' % door.getFullName())

    def onRecordDataUpdated(self, arg):
        """
        Receive RecordDataUpdated tuple, the method detects when the event
        type is 'data_desc' (come with new data description), it will call
        to 'onExpConfChanged' to refresh the plots configuration,
        adding/removing plots based in the new Experimental configuration.

        Note: After the plots reorder, the data description event is resend
        to reconfig the plot with the new data description.

        :param arg: RecordData Tuple
        :return:
        """
        data = arg[1]
        if 'type' in data:
            event_type = data['type']
            if event_type == 'data_desc':
                self.prepare(data)
            elif event_type == 'record_data':
                self.newPoint(data)
            elif event_type == 'record_end':
                self.end(data)

    def prepare(self, data_desc):
        """
        Prepare UI for a new scan. Rebuilds plots as necessary to adapt to the
        new scan channels and moveables
        """
        data = data_desc['data']
        col_map = {column['name']: column for column in data['column_desc']}

        # Build list of curves. Each curve has the same column info from the
        # event and additionally the x-axis it should be plot against
        curves = []
        for column in data['column_desc']:
            ch_name = column['name']
            ptype = column.get('plot_type', PlotType.No)
            if ptype == PlotType.No:
                continue
            elif ptype == PlotType.Image:
                self.warning('Unsupported image plot for %s', ch_name)
                continue
            x_axes = ['point_nb' if axis == '<idx>' else axis
                      for axis in column.get('plot_axes', ())]
            ndim = column.get('ndim', 0) or 0
            if ptype == PlotType.Spectrum:
                if ndim == 0:  # this is a trend
                    for x_axis in x_axes:
                        x_label = col_map[x_axis]['label']
                        curve = dict(column, x_axis=x_axis,
                                     x_axis_label=x_label)
                        curves.append(curve)
                else:
                    self.warning('Cannot create spectrum plot for %d dims '
                                 'channel %r', ndim, ch_name)

        if self._group_mode == self.Single:
            plots = []
            for curve in curves:
                plot = dict(x_axis=dict(name=curve['x_axis'],
                                        label=curve['x_axis_label']),
                            curves=[curve])
                plots.append(plot)
        elif self._group_mode == self.XAxis:
            plot_map = {}
            for curve in curves:
                x_axis = curve['x_axis']
                plot = plot_map.get(x_axis)
                if plot is None:
                    plot = dict(x_axis=dict(name=curve['x_axis'],
                                            label=curve['x_axis_label']),
                                curves=[])
                    plot_map[x_axis] = plot
                plot['curves'].append(curve)
            plots = tuple(plot_map.values())
        else:
            raise NotImplementedError

        nb_points = data.get('total_scan_intervals', 2**16 - 1) + 1
        self._plot.prepare(plots, nb_points=nb_points)

        # build status message
        serialno = 'Scan #{}'.format(data.get('serialno', '?'))
        title = data.get('title', 'unnamed operation')
        if data.get('scandir') and data.get('scanfile'):
            scan_file = data['scanfile']
            if isinstance(scan_file, (list, tuple)):
                scan_file = '&'.join(data['scanfile'])
            saving = data['scandir'] + '/' + scan_file
        else:
            saving = 'no saving!'
        started = 'Started ' + data.get('starttime', '?')
        progress = '{progress}'
        self.message_template = ' | '.join((serialno, title, started,
                                            progress, saving))
        self.newShortMessage.emit(
            self.message_template.format(progress='Preparing...'))

    def newPoint(self, point):
        data = point['data']
        self._plot.onNewPoint(data)
        point_nb = 'Point #{}'.format(data['point_nb'])
        msg = self.message_template.format(progress=point_nb)
        self.newShortMessage.emit(msg)

    def end(self, end_data):
        data = end_data['data']
        self._plot.onEnd(data)
        progress = 'Ended {}'.format(data['endtime'])
        msg = self.message_template.format(progress=progress)
        self.newShortMessage.emit(msg)

    def createPanel(self, widget, name, **kwargs):
        '''Creates a "panel" from a widget. In this basic implementation this
        means that the widgets is shown as a non-modal top window

        :param widget: (QWidget) widget to be used for the panel
        :param name: (str) name of the panel. Must be unique.

        Note: for backawards compatibility, this implementation accepts
        arbitrary keyword arguments which are just ignored
        '''
        widget.setWindowTitle(name)
        widget.show()
        self.__panels[name] = widget

    def getPanelWidget(self, name):
        '''Returns the widget associated to a panel name

        :param name: (str) name of the panel. KeyError is raised if not found

        :return: (QWidget)
        '''
        return self.__panels[name]

    def removePanel(self, name):
        '''stop managing the given panel

        :param name: (str) name of the panel'''
        widget = self.__panels.pop(name)
        if hasattr(widget, 'setModel'):
            widget.setModel(None)
        widget.setParent(None)
        widget.close()


class MacroBroker(DynamicPlotManager):
    '''A manager of all macro-related panels of a TaurusGui.

    It creates, destroys and manages connections for the following objects:

        - Macro Configuration dialog
        - Experiment Configuration panel
        - Macro Executor panel
        - Sequencer panel
        - Macro description viewer
        - Door output, result and debug panels
        - Macro editor
        - Macro "panic" button (to abort macros)
        - Dynamic plots (see :class:`DynamicPlotManager`)
    '''

    def __init__(self, parent):
        '''Passing the parent object (the main window) is mandatory'''
        DynamicPlotManager.__init__(self, parent)

        self._createPermanentPanels()

        # connect the broker to shared data
        Qt.qApp.SDM.connectReader("doorName", self.setModel)

    def setModel(self, doorname):
        ''' Reimplemented from :class:`DynamicPlotManager`.'''
        # disconnect the previous door
        door = self.getModelObj()
        if door is not None:  # disconnect it from *all* shared data providing
            SDM = Qt.qApp.SDM
            try:
                SDM.disconnectWriter("macroStatus", door,
                                     "macroStatusUpdated")
            except Exception:
                self.info("Could not disconnect macroStatusUpdated")
            try:
                SDM.disconnectWriter("doorOutputChanged", door,
                                     "outputUpdated")
            except Exception:
                self.info("Could not disconnect outputUpdated")
            try:
                SDM.disconnectWriter("doorInfoChanged", door, "infoUpdated")
            except Exception:
                self.info("Could not disconnect infoUpdated")
            try:
                SDM.disconnectWriter("doorWarningChanged", door,
                                     "warningUpdated")
            except Exception:
                self.info("Could not disconnect warningUpdated")
            try:
                SDM.disconnectWriter("doorErrorChanged", door, "errorUpdated")
            except Exception:
                self.info("Could not disconnect errorUpdated")
            try:
                SDM.disconnectWriter("doorDebugChanged", door, "debugUpdated")
            except Exception:
                self.info("Could not disconnect debugUpdated")
            try:
                SDM.disconnectWriter("doorResultChanged", door,
                                     "resultUpdated")
            except Exception:
                self.info("Could not disconnect resultUpdated")
            try:
                SDM.disconnectWriter("expConfChanged", door,
                                     "experimentConfigurationChanged")
            except Exception:
                self.info(
                    "Could not disconnect experimentConfigurationChanged")
        # set the model
        DynamicPlotManager.setModel(self, doorname)

        # connect the new door
        door = self.getModelObj()
        if door is not None:
            SDM = Qt.qApp.SDM
            SDM.connectWriter("macroStatus", door, "macroStatusUpdated")
            SDM.connectWriter("doorOutputChanged", door, "outputUpdated")
            SDM.connectWriter("doorInfoChanged", door, "infoUpdated")
            SDM.connectWriter("doorWarningChanged", door, "warningUpdated")
            SDM.connectWriter("doorErrorChanged", door, "errorUpdated")
            SDM.connectWriter("doorDebugChanged", door, "debugUpdated")
            SDM.connectWriter("doorResultChanged", door, "resultUpdated")
            SDM.connectWriter("expConfChanged", door,
                              "experimentConfigurationChanged")

    def _createPermanentPanels(self):
        '''creates panels on the main window'''
        from sardana.taurus.qt.qtgui.extra_macroexecutor import \
            TaurusMacroExecutorWidget, TaurusSequencerWidget, \
            TaurusMacroConfigurationDialog, TaurusMacroDescriptionViewer, \
            DoorOutput, DoorDebug, DoorResult

        from sardana.taurus.qt.qtgui.extra_sardana import \
            ExpDescriptionEditor

        mainwindow = self.parent()

        # Create macroconfiguration dialog & action
        self.__macroConfigurationDialog = \
            TaurusMacroConfigurationDialog(mainwindow)
        self.macroConfigurationAction = mainwindow.taurusMenu.addAction(
            Qt.QIcon.fromTheme("preferences-system-session"),
            "Macro execution configuration...",
            self.__macroConfigurationDialog.show)

        SDM = Qt.qApp.SDM
        SDM.connectReader("macroserverName",
                          self.__macroConfigurationDialog.selectMacroServer)
        SDM.connectReader("doorName",
                          self.__macroConfigurationDialog.selectDoor)
        SDM.connectWriter("macroserverName", self.__macroConfigurationDialog,
                          'macroserverNameChanged')
        SDM.connectWriter("doorName", self.__macroConfigurationDialog,
                          'doorNameChanged')

        # Create ExpDescriptionEditor dialog
        self.__expDescriptionEditor = ExpDescriptionEditor()
        SDM.connectReader("doorName", self.__expDescriptionEditor.setModel)
        mainwindow.createPanel(self.__expDescriptionEditor,
                               'Experiment Config',
                               registerconfig=True,
                               icon=Qt.QIcon.fromTheme('preferences-system'),
                               permanent=True)
        ###############################
        # TODO: These lines can be removed once the door does emit
        # "experimentConfigurationChanged" signals
        SDM.connectWriter("expConfChanged", self.__expDescriptionEditor,
                          "experimentConfigurationChanged")
        ################################

        # put a Macro Executor
        self.__macroExecutor = TaurusMacroExecutorWidget()
        SDM.connectReader("macroserverName", self.__macroExecutor.setModel)
        SDM.connectReader("doorName", self.__macroExecutor.onDoorChanged)
        SDM.connectReader("macroStatus",
                          self.__macroExecutor.onMacroStatusUpdated)
        SDM.connectWriter("macroName", self.__macroExecutor,
                          "macroNameChanged")
        SDM.connectWriter("executionStarted", self.__macroExecutor,
                          "macroStarted")
        SDM.connectWriter("plotablesFilter", self.__macroExecutor,
                          "plotablesFilterChanged")
        SDM.connectWriter("shortMessage", self.__macroExecutor,
                          "shortMessageEmitted")
        mainwindow.createPanel(self.__macroExecutor, 'Macros',
                               registerconfig=True, permanent=True)

        # put a Sequencer
        self.__sequencer = TaurusSequencerWidget()
        SDM.connectReader("macroserverName", self.__sequencer.setModel)
        SDM.connectReader("doorName", self.__sequencer.onDoorChanged)
        SDM.connectReader("macroStatus",
                          self.__sequencer.onMacroStatusUpdated)
        SDM.connectWriter("macroName", self.__sequencer.tree,
                          "macroNameChanged")
        SDM.connectWriter("macroName", self.__sequencer,
                          "macroNameChanged")
        SDM.connectWriter("executionStarted", self.__sequencer,
                          "macroStarted")
        SDM.connectWriter("plotablesFilter", self.__sequencer,
                          "plotablesFilterChanged")
        SDM.connectWriter("shortMessage", self.__sequencer,
                          "shortMessageEmitted")
        mainwindow.createPanel(self.__sequencer, 'Sequences',
                               registerconfig=True, permanent=True)

        # puts a macrodescriptionviewer
        self.__macroDescriptionViewer = TaurusMacroDescriptionViewer()
        SDM.connectReader("macroserverName",
                          self.__macroDescriptionViewer.setModel)
        SDM.connectReader("macroName",
                          self.__macroDescriptionViewer.onMacroNameChanged)
        mainwindow.createPanel(self.__macroDescriptionViewer,
                               'MacroDescription', registerconfig=True,
                               permanent=True)

        # puts a doorOutput
        self.__doorOutput = DoorOutput()
        SDM.connectReader("doorOutputChanged",
                          self.__doorOutput.onDoorOutputChanged)
        SDM.connectReader("doorInfoChanged",
                          self.__doorOutput.onDoorInfoChanged)
        SDM.connectReader("doorWarningChanged",
                          self.__doorOutput.onDoorWarningChanged)
        SDM.connectReader("doorErrorChanged",
                          self.__doorOutput.onDoorErrorChanged)
        SDM.connectReader("doorDebugChanged",
                          self.__doorOutput.onDoorDebugChanged)
        mainwindow.createPanel(self.__doorOutput, 'DoorOutput',
                               registerconfig=False, permanent=True)

        # puts doorResult
        self.__doorResult = DoorResult(mainwindow)
        SDM.connectReader("doorResultChanged",
                          self.__doorResult.onDoorResultChanged)
        mainwindow.createPanel(self.__doorResult, 'DoorResult',
                               registerconfig=False, permanent=True)

        # puts sardanaEditor
        # self.__sardanaEditor = SardanaEditor()
        # SDM.connectReader("macroserverName", self.__sardanaEditor.setModel)
        # mainwindow.createPanel(self.__sardanaEditor, 'SardanaEditor',
        #                        registerconfig=False, permanent=True)

        # add panic button for aborting the door
        text = "Panic Button: stops the pool (double-click for abort)"
        self.doorAbortAction = mainwindow.jorgsBar.addAction(
            Qt.QIcon("actions:process-stop.svg"),
            text, self.__onDoorAbort)

        # store beginning of times as a datetime
        self.__lastAbortTime = datetime.datetime(1, 1, 1)

        # store doubleclick interval as a timedelta
        td = datetime.timedelta(0, 0, 1000 * Qt.qApp.doubleClickInterval())
        self.__doubleclickInterval = td

    def __onDoorAbort(self):
        '''slot to be called when the abort action is triggered.
        It sends stop command to the pools (or abort if the action
        has been triggered twice in less than self.__doubleclickInterval

        .. note:: An abort command is always preceded by an stop command
        '''
        # decide whether to send stop or abort
        now = datetime.datetime.now()
        if now - self.__lastAbortTime < self.__doubleclickInterval:
            cmd = 'abort'
        else:
            cmd = 'stop'

        door = self.getModelObj()

        # abort the door
        door.command_inout('abort')
        # send stop/abort to all pools
        pools = door.macro_server.getElementsOfType('Pool')
        for pool in list(pools.values()):
            self.info('Sending %s command to %s' % (cmd, pool.getFullName()))
            try:
                pool.getObj().command_inout(cmd)
            except Exception:
                self.info('%s command failed on %s', cmd, pool.getFullName(),
                          exc_info=1)
        self.newShortMessage.emit("%s command sent to all pools" % cmd)
        self.__lastAbortTime = now

    def createPanel(self, widget, name, **kwargs):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        return mainwindow.createPanel(widget, name, **kwargs)

    def getPanelWidget(self, name):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        return mainwindow.getPanel(name).widget()

    def removePanel(self, name):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        mainwindow.removePanel(name)

    def removeTemporaryPanels(self, names=None):
        '''Remove temporary panels managed by this widget'''
        # for now, the only temporary panels are the plots
        DynamicPlotManager.removePanels(self, names=names)


if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication()

    b = DynamicPlotManager(None)

    b.setModel('door/cp1/1')

    print('...')
    sys.exit(app.exec_())



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__all__ = ["PseudoMotor", "PseudoMotorClass"]

__docformat__ = 'restructuredtext'

import sys
import time

from PyTango import DevFailed, Except, READ_WRITE, SCALAR, DevVoid, \
    DevDouble, DevBoolean, DevVarStringArray, DevVarDoubleArray, DevState, \
    AttrQuality

from taurus.core.util.log import DebugIt

from sardana import State, SardanaServer
from sardana.sardanaexception import SardanaException
from sardana.sardanaattribute import SardanaAttribute
from sardana.tango.core.util import exception_str, to_tango_type_format, \
    throw_sardana_exception
from sardana.tango.pool.PoolDevice import PoolElementDevice, \
    PoolElementDeviceClass


class PseudoMotor(PoolElementDevice):

    def __init__(self, dclass, name):
        self.in_write_position = False
        PoolElementDevice.__init__(self, dclass, name)

    def init(self, name):
        PoolElementDevice.init(self, name)

    def _is_allowed(self, req_type):
        return PoolElementDevice._is_allowed(self, req_type)

    def get_pseudo_motor(self):
        return self.element

    def set_pseudo_motor(self, pseudo_motor):
        self.element = pseudo_motor

    pseudo_motor = property(get_pseudo_motor, set_pseudo_motor)

    @DebugIt()
    def delete_device(self):
        PoolElementDevice.delete_device(self)
        pseudo_motor = self.pseudo_motor
        if pseudo_motor is not None:
            pseudo_motor.remove_listener(self.on_pseudo_motor_changed)

    @DebugIt()
    def init_device(self):
        PoolElementDevice.init_device(self)

        self.Elements = list(map(int, self.Elements))
        pseudo_motor = self.pseudo_motor
        if self.pseudo_motor is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.pseudo_motor = pseudo_motor = \
                self.pool.create_element(type="PseudoMotor", name=name,
                                         full_name=full_name, id=self.Id, axis=self.Axis,
                                         ctrl_id=self.Ctrl_id, user_elements=self.Elements)
            if self.instrument is not None:
                pseudo_motor.set_instrument(self.instrument)
        pseudo_motor.set_drift_correction(self.DriftCorrection)
        pseudo_motor.add_listener(self.on_pseudo_motor_changed)

        self.set_state(DevState.ON)

    def on_pseudo_motor_changed(self, event_source, event_type, event_value):
        try:
            self._on_pseudo_motor_changed(event_source, event_type,
                                          event_value)
        except:
            msg = 'Error occurred "on_pseudo_motor_changed(%s.%s): %s"'
            exc_info = sys.exc_info()
            self.error(msg, self.pseudo_motor.name, event_type.name,
                       exception_str(*exc_info[:2]))
            self.debug("Details", exc_info=exc_info)

    def _on_pseudo_motor_changed(self, event_source, event_type, event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name.lower()

        try:
            attr = self.get_attribute_by_name(name)
        except DevFailed:
            return
        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        value, w_value, error = None, None, None

        if name == "state":
            value = self.calculate_tango_state(event_value)
        elif name == "status":
            value = self.calculate_tango_status(event_value)
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                else:
                    value = event_value.value
                timestamp = event_value.timestamp
            else:
                value = event_value
            state = self.pseudo_motor.get_state(propagate=0)

            if name == "position":
                w_value = event_value.w_value
                if state == State.Moving:
                    quality = AttrQuality.ATTR_CHANGING

        self.set_attribute(attr, value=value, w_value=w_value,
                           timestamp=timestamp, quality=quality,
                           priority=priority, error=error, synch=False)

    def always_executed_hook(self):
        #state = to_tango_state(self.pseudo_motor.get_state(cache=False))
        pass

    def read_attr_hardware(self, data):
        pass

    def get_dynamic_attributes(self):
        cache_built = hasattr(self, "_dynamic_attributes_cache")

        std_attrs, dyn_attrs = \
            PoolElementDevice.get_dynamic_attributes(self)

        if not cache_built:
            # For position attribute, listen to what the controller says for
            # data type (between long and float)
            pos = std_attrs.get('position')
            if pos is not None:
                _, data_info, attr_info = pos
                ttype, _ = to_tango_type_format(attr_info.dtype)
                data_info[0][0] = ttype
        return std_attrs, dyn_attrs

    def initialize_dynamic_attributes(self):
        attrs = PoolElementDevice.initialize_dynamic_attributes(self)

        detect_evts = "position",
        non_detect_evts = ()

        for attr_name in detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, True)
        for attr_name in non_detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, False)
        return

    def read_Position(self, attr):
        pseudo_motor = self.pseudo_motor
        use_cache = pseudo_motor.is_in_operation() and not self.Force_HW_Read
        position = pseudo_motor.get_position(cache=use_cache, propagate=0)
        if position.error:
            throw_sardana_exception(position)
        state = pseudo_motor.get_state(cache=use_cache, propagate=0)
        quality = None
        if state == State.Moving:
            quality = AttrQuality.ATTR_CHANGING
        self.set_attribute(attr, value=position.value, w_value=position.w_value,
                           quality=quality, priority=0,
                           timestamp=position.timestamp)

    def write_Position(self, attr):
        self.in_write_position = True
        try:
            position = attr.get_write_value()
            self.debug("write_Position(%s)", position)
            try:
                self.wait_for_operation()
            except:
                raise Exception("Cannot move: already in motion")
            pseudo_motor = self.pseudo_motor
            try:
                pseudo_motor.position = position
            except SardanaException as se:
                throw_sardana_exception(se)
        finally:
            self.in_write_position = False

    def CalcPseudo(self, physical_positions):
        """Returns the pseudo motor position for the given physical positions"""
        if not len(physical_positions):
            physical_positions = None
        result = self.pseudo_motor.calc_pseudo(
            physical_positions=physical_positions)
        if result.error:
            throw_sardana_exception(result)
        return result.value

    def CalcPhysical(self, pseudo_position):
        """Returns the physical motor positions for the given pseudo motor
        position assuming the current pseudo motor write positions for all the
        other sibling pseudo motors"""
        result = self.pseudo_motor.calc_physical(pseudo_position)
        if result.error:
            throw_sardana_exception(result)
        return result.value

    def CalcAllPhysical(self, pseudo_positions):
        """Returns the physical motor positions for the given pseudo motor
        position(s)"""
        result = self.pseudo_motor.calc_physical(pseudo_positions)
        if result.error:
            throw_sardana_exception(result)
        return result.value

    def CalcAllPseudo(self, physical_positions):
        """Returns the pseudo motor position(s) for the given physical positions"""
        result = self.pseudo_motor.calc_all_pseudo(physical_positions)
        if result.error:
            throw_sardana_exception(result)
        return result.value

    def MoveRelative(self, argin):
        raise NotImplementedError

    def is_MoveRelative_allowed(self):
        if self.get_state() in (DevState.FAULT, DevState.MOVING,
                                DevState.UNKNOWN):
            return False
        return True

    is_Position_allowed = _is_allowed


class PseudoMotorClass(PoolElementDeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
        "Elements":    [DevVarStringArray, "elements used by the pseudo", []],
        'DriftCorrection':
            [DevBoolean,
             "Locally apply drift correction on pseudo motors. Default is the "
             "current global drift correction in the Pool Device",
             None],
    }
    device_property_list.update(PoolElementDeviceClass.device_property_list)

    #    Command definitions
    cmd_list = {
        'CalcPseudo': [[DevVarDoubleArray, "physical positions"], [DevDouble, "pseudo position"]],
        'CalcPhysical': [[DevDouble, "pseudo position"], [DevVarDoubleArray, "physical positions"]],
        'CalcAllPseudo': [[DevVarDoubleArray, "physical positions"], [DevVarDoubleArray, "pseudo positions"]],
        'CalcAllPhysical': [[DevVarDoubleArray, "pseudo positions"], [DevVarDoubleArray, "physical positions"]],
        'MoveRelative': [[DevDouble, "amount to move"], [DevVoid, ""]],
    }
    cmd_list.update(PoolElementDeviceClass.cmd_list)

    #    Attribute definitions
    standard_attr_list = {
        'Position': [[DevDouble, SCALAR, READ_WRITE],
                     {'label': "Position",
                      'abs_change': '1.0', }, ],
    }
    standard_attr_list.update(PoolElementDeviceClass.standard_attr_list)

    def _get_class_properties(self):
        ret = PoolElementDeviceClass._get_class_properties(self)
        ret['Description'] = "Pseudo motor device class"
        ret['InheritedFrom'].insert(0, 'PoolElementDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Tests for expert macros"""

import unittest
from sardana.macroserver.macros.test import RunMacroTestCase, testRun, getCTs,\
    getControllers

CTRL_NAME1 = getControllers()[0]
CT_NAME1, CT_NAME2 = getCTs()[:2]


class ExpertTest(RunMacroTestCase, unittest.TestCase):
    """Test case for some of the expert macros.
    """

    def test_expert(self):
        CTRL_NAME = "unittestmotctrl01"
        MOT_NAME1 = "unittestmot01"
        MOT_NAME2 = "unittestmot02"
        try:
            self.macro_runs(macro_name="defctrl",
                            macro_params=["DummyMotorController", CTRL_NAME],
                            wait_timeout=1)
            self.macro_runs(macro_name="defelem",
                            macro_params=[MOT_NAME1, CTRL_NAME, "1"],
                            wait_timeout=1)
            self.macro_runs(macro_name="renameelem",
                            macro_params=[MOT_NAME1, MOT_NAME2],
                            wait_timeout=1)
            self.macro_runs(macro_name="defm",
                            macro_params=[MOT_NAME1, CTRL_NAME, "2"],
                            wait_timeout=1)
            self.macro_runs(macro_name="udefelem",
                            macro_params=[MOT_NAME1, MOT_NAME2],
                            wait_timeout=1)
            self.macro_runs(macro_name="udefctrl",
                            macro_params=[CTRL_NAME],
                            wait_timeout=1)
        except Exception as e:
            import taurus
            taurus.warning("Your system may stay dirty due to an unexpected"
                           " exception during the test.")
            raise e


class MeasTest(RunMacroTestCase, unittest.TestCase):
    """Test case for measurement group related expert macros.
    """
    @unittest.skip("skip test until issue #57 is fixed")
    def test_meas(self):
        MNTGRP_NAME = "unittestmntgrp01"
        try:
            self.macro_runs(macro_name="defmeas",
                            macro_params=[MNTGRP_NAME, CT_NAME1, CT_NAME2])
            self.macro_runs(macro_name="udefmeas",
                            macro_params=[MNTGRP_NAME])
        except Exception as e:
            import taurus
            taurus.warning("Your system may stay dirty due to an unexpected"
                           " exception during the test.")
            raise e

# TODO: improve this test: not all sardana controller implement SendToCtrl
# @testRun(macro_params=[CTRL_NAME1, "blabla"], wait_timeout=1)
# class Send2ctrlTest(RunMacroTestCase, unittest.TestCase):
#     """Test case for send2ctrl macro.
#     """
#     macro_name = "send2ctrl"


# This is a known failure until bug-472 is fixed:
# https://sourceforge.net/p/sardana/tickets/472/
# @testRun(macro_params=["DummyMotorController"])
# class EdctrlTest(RunMacroTestCase, unittest.TestCase):
#     """Test case for edctrl macro.
#     """
#     macro_name = "edctrl"


@testRun(macro_name="prdef", macro_params=["wa"], wait_timeout=1)
@testRun(macro_name="relmaclib", macro_params=["standard"], wait_timeout=1)
@testRun(macro_name="relmac", macro_params=["wa"], wait_timeout=1)
class MacroTest(RunMacroTestCase, unittest.TestCase):
    """Test case for macro related expert macros.
    """


@testRun(macro_params=["wa"], wait_timeout=1)
@testRun(macro_params=[CTRL_NAME1], wait_timeout=1)
@testRun(macro_params=[CT_NAME1], wait_timeout=1)
class SarinfoTest(RunMacroTestCase, unittest.TestCase):
    """Test case for send2ctrl macro.
    """
    macro_name = "sar_info"



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the class definition for the MacroServer macro
manager"""

__all__ = ["RecorderManager"]

__docformat__ = 'restructuredtext'

import os
import sys
import copy
import inspect

from collections import OrderedDict

from sardana import sardanacustomsettings
from sardana.sardanaexception import format_exception_only_str
from sardana.sardanamodulemanager import ModuleManager
from sardana.macroserver.msmanager import MacroServerManager
from sardana.macroserver.scan.recorder import DataRecorder
from sardana.macroserver.msmetarecorder import RecorderLibrary, \
    RecorderClass
from sardana.macroserver.msexception import UnknownRecorder, LibraryError

_BASE_DIR = os.path.dirname(os.path.abspath(__file__))


class RecorderManager(MacroServerManager):

    DEFAULT_RECORDER_DIRECTORIES = os.path.join(_BASE_DIR, 'recorders'),

    def __init__(self, macro_server, recorder_path=None):
        MacroServerManager.__init__(self, macro_server)
        if recorder_path is not None:
            self.setRecorderPath(recorder_path)

    def reInit(self):
        if self.is_initialized():
            return

        # dict<str, RecorderLibrary>
        # key   - module name (without path and without extension)
        # value - RecorderLibrary object representing the module
        self._modules = {}

        #: dict<str, <metarecorder.RecorderClass>
        #: key   - recorder name
        #: value - RecorderClass object representing the recorder
        self._recorder_dict = {}

        # list<str>
        # elements are absolute paths
        self._recorder_path = []

        # custom map (per installation) allowing to avoid
        # recorder class ambiguity problems (using extension filter)
        #: dict<str, str>
        #: key   - scan file extension
        #: value - recorder name
        self._custom_scan_recorder_map = getattr(sardanacustomsettings,
                                                 "SCAN_RECORDER_MAP",
                                                 {})
        #: dict<str, str>
        #: key   - scan file extension
        #: value - list with recorder name(s)
        self._scan_recorder_map = {}

        MacroServerManager.reInit(self)

    def setScanRecorderMap(self, recorder_map):
        """Registers a new map of recorders in this manager.
        """
        self._scan_recorder_map = dict(recorder_map)

    def setRecorderPath(self, recorder_path):
        """Registers a new list of recorder directories in this manager.
        """
        _recorder_path = []
        for paths in recorder_path:
            splited_paths = paths.split(os.pathsep)
            for path in splited_paths:
                # filter empty and commented paths
                if not path.startswith("#"):
                    _recorder_path.append(path)

        for recorder_dir in self.DEFAULT_RECORDER_DIRECTORIES:
            if recorder_dir not in _recorder_path:
                _recorder_path.append(recorder_dir)

        self._recorder_path = _recorder_path

        recorder_file_names = self._findRecorderLibNames(
            _recorder_path)

        for mod_name, file_name in recorder_file_names.items():
            dir_name = os.path.dirname(file_name)
            path = [dir_name]
            try:
                self.reloadRecorderLib(mod_name, path)
            except:
                pass
        pass

    def getRecorderPath(self):
        return self._recorder_path

    def getRecorderMetaClass(self, recorder_name):
        """ Return the Recorder class for the given class name.
        :param klass_name: Name of the recorder class.
        :type klass_name: :obj:`str`
        :return:  a :obj:`class` class of recorder or None if it does not exist
        :rtype:
            :obj:`class:`~sardana.macroserver.msmetarecorder.RecorderClass`\>
        """
        ret = self._recorder_dict.get(recorder_name)
        if ret is None:
            raise UnknownRecorder("Unknown recorder %s" % recorder_name)
        return ret

    def getRecorderMetaClasses(self, filter=None, extension=None):
        """ Returns a :obj:`dict` containing information about recorder
        classes. These may be limitted by two conditions - filter and
        extension. The first one selects just the classes inheriting from the
        filter. The second one selects just the classes implementing a given
        extension (format). Both can be used at the same time.

        :param filter: a klass of a valid type of Recorder
        :type filter: obj
        :param filter: a scan file extension
        :type filter: :obj:`str`
        :return: a :obj:`dict` containing information about recorder classes
        :rtype:
            :obj:`dict`\<:obj:`str`\,
            :class:`~sardana.macroserver.msmetarecorder.RecorderClass`\>
        """
        if filter is None:
            filter = DataRecorder
        ret = {}
        for name, klass in list(self._recorder_dict.items()):
            if not issubclass(klass.recorder_class, filter):
                continue
            if extension is not None:
                # fist look into the SCAN_RECORDER_MAP
                _map = self._custom_scan_recorder_map or {}
                name = _map.get(extension, None)
                if name is not None:
                    klass = self.getRecorderMetaClass(name)
                # second look into the standard map
                else:
                    _map = self._scan_recorder_map
                    if extension not in list(_map.keys()):
                        continue
                    elif klass not in _map[extension]:
                        continue
            ret[name] = klass
        return ret

    def getRecorderClasses(self, filter=None, extension=None):
        """ Returns a :obj:`dict` containing information about recorder classes.
        :param filter: a klass of a valid type of Recorder
        :type filter: obj
        :param filter: a scan file extension
        :type filter: :obj:`str`
        :return: a :obj:`dict` containing information about recorder classes
        :rtype:
            :obj:`dict`\<:obj:`str`\, :class:`DataRecorder`\>
        """
        if filter is None:
            filter = DataRecorder
        meta_klasses = self.getRecorderMetaClasses(filter=filter,
                                                   extension=extension)
        return dict((key, value.klass)
                    for (key, value) in list(meta_klasses.items()))

    def getRecorderClass(self, klass_name):
        """ Return the Recorder class for the given class name.
        :param klass_name: Name of the recorder class.
        :type klass_name: :obj:`str`
        :return:  a :obj:`class` class of recorder or None if it does not exist
        :rtype:
            :obj:`class:`DataRecorder`\>
        """
        meta_klass = self.getRecorderMetaClass(klass_name)
        return meta_klass.klass

    def _findRecorderLibName(self, lib_name, path=None):
        path = path or self.getRecorderPath()
        f_name = lib_name
        if not f_name.endswith('.py'):
            f_name += '.py'
        for p in path:
            try:
                elems = os.listdir(p)
                if f_name in elems:
                    return os.path.abspath(os.path.join(p, f_name))
            except:
                self.debug("'%s' is not a valid path" % p)
        return None

    def _findRecorderLibNames(self, recorder_path=None):
        path = recorder_path or self.getRecorderPath()
        ret = OrderedDict()
        for p in reversed(path):
            try:
                for fdir in os.listdir(p):
                    name, ext = os.path.splitext(fdir)
                    if name.startswith("_"):
                        continue
                    if ext.endswith('.py'):
                        full_path = os.path.abspath(os.path.join(p, fdir))
                        ret[name] = full_path
            except:
                self.debug("'%s' is not a valid path" % p)
        return ret

    def reloadRecorderLib(self, module_name, path=None):
        """Reloads the given library(=module) names.

        :param module_name: recorder library name (=python module name)
        :param path:
            a list of absolute path to search for libraries [default: None,
            means the current RecorderPath will be used]"""
        path = path or self.getRecorderPath()
        # reverse the path order:
        # more priority elements last. This way if there are repeated elements
        # they first ones (lower priority) will be overwritten by the last ones
        if path:
            path = copy.copy(path)
            path.reverse()

        # if there was previous Recorder Library info remove it
        old_recorder_lib = self._modules.pop(module_name, None)
        if old_recorder_lib is not None:
            for recorder in old_recorder_lib.get_recorders():
                self._recorder_dict.pop(recorder.name)
                # remove recorders from the map
                for _, recorders in self._scan_recorder_map.items():
                    try:
                        recorders.remove(recorder)
                    except:
                        pass

        mod_manager = ModuleManager()
        m, exc_info = None, None
        try:
            m = mod_manager.reloadModule(module_name, path)
        except:
            exc_info = sys.exc_info()

        params = dict(module=m, name=module_name,
                      macro_server=self.macro_server, exc_info=exc_info)
        if m is None:
            file_name = self._findRecorderLibName(module_name)
            if file_name is None:
                if exc_info:
                    msg = format_exception_only_str(*exc_info[:2])
                else:
                    msg = "Error (re)loading recorder library '%s'" \
                        % module_name
                raise LibraryError(msg, exc_info=exc_info)
            params['file_path'] = file_name
            recorder_lib = RecorderLibrary(**params)
            self._modules[module_name] = recorder_lib
        else:
            recorder_lib = RecorderLibrary(**params)
            lib_contains_recorders = False
            abs_file = recorder_lib.file_path
            for _, klass in inspect.getmembers(m, inspect.isclass):
                if issubclass(klass, DataRecorder):
                    # if it is a class defined in some other class forget it to
                    # avoid replicating the same recorder in different
                    # recorder files
                    # use normcase to treat case insensitivity of paths on
                    # certain platforms e.g. Windows
                    if os.path.normcase(inspect.getabsfile(klass)) !=\
                       os.path.normcase(abs_file):
                        continue
                    lib_contains_recorders = True
                    self.addRecorder(recorder_lib, klass)
            if lib_contains_recorders:
                self._modules[module_name] = recorder_lib

        return recorder_lib

    def addRecorder(self, recorder_lib, klass):
        """Adds a new recorder class
        """
        recorder_name = klass.__name__
        exists = recorder_lib.has_recorder(recorder_name)
        if exists:
            action = "Updating"
        else:
            action = "Adding"

        self.debug("%s recorder %s" % (action, recorder_name))

        try:
            recorder_class = RecorderClass(lib=recorder_lib, klass=klass,
                                           macro_server=self.macro_server)
            recorder_lib.add_recorder(recorder_class)
            self._recorder_dict[recorder_name] = recorder_class
            if hasattr(klass, "formats"):
                self._addRecorderToMap(recorder_class)
        except:
            self.warning("Failed to add recorder class %s", recorder_name,
                         exc_info=1)

        if exists:
            action = "Updated"
        else:
            action = "Added"
        self.debug("%s recorder %s" % (action, recorder_name))

    def _addRecorderToMap(self, recorder_class):
        klass = recorder_class.klass
        for ext in list(klass.formats.values()):
            recorders = self._scan_recorder_map.get(ext, [])
            if len(recorders) == 0:
                recorders.append(recorder_class)
            else:
                recorder_from_map = recorders[-1]  # it could be any recorder
                # recorders are on the same priority level (located in the same
                # directory) - just append it to the list
                if recorder_from_map.lib.path == recorder_class.lib.path:
                    recorders.append(recorder_class)
                # new recorder comes from another directory (it must be of
                # higher priority) - forget about others and create new list
                else:
                    recorders = [recorder_class]
            self._scan_recorder_map[ext] = recorders
